/*
 * dk.brics.automaton
 * 
 * Copyright (c) 2001-2011 Anders Moeller All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer. 2. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution. 3. The name of the author may not
 * be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package dk.brics.automaton2;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import gnu.trove.set.hash.THashSet;

/**
 * <tt>Automaton</tt> state.
 * 
 * @author Anders M&oslash;ller
 *         &lt;<a href="mailto:amoeller@cs.au.dk">amoeller@cs.au.dk</a>&gt;
 */
public class State implements Serializable, Comparable<State> {

	static final long serialVersionUID = 30001;

	boolean accept;
	Set<Transition> transitions;

	int number;

	final long id;
	static AtomicLong next_id = new AtomicLong(0);
	AtomicInteger lastVisit = new AtomicInteger(0);

	/**
	 * Constructs a new state. Initially, the new state is a reject state.
	 */
	public State() {
		resetTransitions();
		id = next_id.incrementAndGet();
	}

	/**
	 * Resets transition set.
	 */
	final void resetTransitions() {
		transitions = new THashSet<Transition>();
	}

	/**
	 * Returns the set of outgoing transitions. Subsequent changes are reflected
	 * in the automaton.
	 * 
	 * @return transition set
	 */
	public Set<Transition> getTransitions() {
		return transitions;
	}

	/**
	 * Adds an outgoing transition.
	 * 
	 * @param t
	 *            transition
	 */
	public void addTransition(Transition t) {
		transitions.add(t);
	}

	/**
	 * Sets acceptance for this state.
	 * 
	 * @param accept
	 *            if true, this state is an accept state
	 */
	public void setAccept(boolean accept) {
		this.accept = accept;
	}

	/**
	 * Returns acceptance status.
	 * 
	 * @return true is this is an accept state
	 */
	public boolean isAccept() {
		return accept;
	}

	/**
	 * Performs lookup in transitions, assuming determinism.
	 * 
	 * @param c
	 *            character to look up
	 * @return destination state, null if no matching outgoing transition
	 * @see #step(char, Collection)
	 */
	public State step(char c) {
		for (Transition t : transitions)
			if (t.min <= c && c <= t.max)
				return t.to;
		return null;
	}

	/**
	 * Performs lookup in transitions, allowing nondeterminism.
	 * 
	 * @param c
	 *            character to look up
	 * @param dest
	 *            collection where destination states are stored
	 * @see #step(char)
	 */
	public void step(char c, Collection<State> dest) {
		for (Transition t : transitions)
			if (t.min <= c && c <= t.max)
				dest.add(t.to);
	}

	boolean addEpsilon(State to) {
		boolean result = false;
		if (to.accept) {
			result = !accept;
			accept = true;
		}
		return transitions.addAll(to.transitions) || result;
	}

	/**
	 * Returns transitions sorted by (min, reverse max, to) or (to, min, reverse
	 * max)
	 */
	Transition[] getSortedTransitionArray(boolean to_first) {
		Transition[] e = transitions.toArray(new Transition[transitions.size()]);
		Arrays.sort(e, new TransitionComparator(to_first));
		return e;
	}

	/**
	 * Returns sorted list of outgoing transitions.
	 * 
	 * @param to_first
	 *            if true, order by (to, min, reverse max); otherwise (min,
	 *            reverse max, to)
	 * @return transition list
	 */
	public List<Transition> getSortedTransitions(boolean to_first) {
		return Arrays.asList(getSortedTransitionArray(to_first));
	}

	/**
	 * Returns string describing this state. Normally invoked via
	 * {@link Automaton#toString()}.
	 */
	@Override
	public String toString() {
		StringBuilder b = new StringBuilder();
		b.append("state ").append(number).append(" ").append(id);
		if (accept)
			b.append(" [accept]");
		else
			b.append(" [reject]");
		b.append(":\n");
		for (Transition t : transitions)
			b.append("  ").append(t.toString()).append("\n");
		return b.toString();
	}

	/**
	 * Compares this object with the specified object for order. States are
	 * ordered by the time of construction.
	 */
	public int compareTo(State s) {
		return Long.compare(s.id, id);
	}

	/**
	 * See {@link java.lang.Object#equals(java.lang.Object)}.
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof State) {
			return ((State) obj).id == id;
		}
		return super.equals(obj);
	}

	/**
	 * See {@link java.lang.Object#hashCode()}.
	 */
	@Override
	public int hashCode() {
		return (int) (id ^ (id >>> 32));
	}

	/**
	 * 
	 * 
	 * @param visitor
	 * @return whether this visitor already visited this state.
	 * @throws ConcurrentModificationException
	 *             thrown when a later visitor already visited
	 */
	public boolean alreadyVisited(int visitor) throws ConcurrentModificationException {
		while (true) {
			int local = lastVisit.get();
			if (visitor == local) {
				return true;
			} else if (visitor < local) {
				System.out.println("multiple visits at the same time");
				throw new ConcurrentModificationException(); // there was already a next visitor here
			} else if (lastVisit.compareAndSet(local, visitor)) {
				return false; // swap successful
			}
			// keep trying
		}
	}

	/**
	 * 
	 * @return the id of this state. Guaranteed to be unique on a single machine
	 *         (using AtomicLong).
	 */
	public long getId() {
		return id;
	}
}
