/***********************************************************
 * This software is part of the ProM package * http://www.processmining.org/ * *
 * Copyright (c) 2003-2008 TU/e Eindhoven * and is licensed under the * LGPL
 * License, Version 1.0 * by Eindhoven University of Technology * Department of
 * Information Systems * http://www.processmining.org * *
 ***********************************************************/

package org.processmining.plugins.petrinet.behavioralanalysis;

import java.util.HashSet;
import java.util.Set;

import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.connections.ConnectionManager;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.AbstractSemanticConnection;
import org.processmining.models.connections.petrinets.behavioral.HomeMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.StateSpaceConnection;
import org.processmining.models.graphbased.directed.analysis.ShortestPathFactory;
import org.processmining.models.graphbased.directed.analysis.ShortestPathInfo;
import org.processmining.models.graphbased.directed.petrinet.InhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.ResetInhibitorNet;
import org.processmining.models.graphbased.directed.petrinet.ResetNet;
import org.processmining.models.graphbased.directed.petrinet.analysis.HomeMarkingSet;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.models.graphbased.directed.transitionsystem.ReachabilityGraph;
import org.processmining.models.graphbased.directed.transitionsystem.State;
import org.processmining.models.semantics.Semantics;
import org.processmining.models.semantics.petrinet.InhibitorNetSemantics;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.models.semantics.petrinet.PetrinetSemantics;
import org.processmining.models.semantics.petrinet.ResetInhibitorNetSemantics;
import org.processmining.models.semantics.petrinet.ResetNetSemantics;
import org.processmining.models.semantics.petrinet.impl.PetrinetSemanticsFactory;

/**
 * Class to analyze all home marking Home marking is reachable from all of
 * available marking. This class can only analyze bounded net
 * 
 * @author arya
 * @email arya.adriansyah@gmail.com
 * @version Dec 13, 2008
 */
@Plugin(name = "Analyze Home Marking", returnLabels = { "Home Marking info" }, returnTypes = { HomeMarkingSet.class }, parameterLabels = {
		"Net", "Marking", "Reachability Graph", "Semantic" }, userAccessible = true)
public class HomeMarkingAnalyzer {
	/**
	 * Variant of net and marking
	 */
	// variant with Petrinet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, Petrinet net, Marking state)
			throws ConnectionCannotBeObtained {
		return analyzeHomeMarkingPetriNet(context, net, state, PetrinetSemanticsFactory
				.regularPetrinetSemantics(Petrinet.class));
	}

	// variant with InhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, InhibitorNet net, Marking state)
			throws ConnectionCannotBeObtained {
		return analyzeHomeMarkingPetriNet(context, net, state, PetrinetSemanticsFactory
				.regularInhibitorNetSemantics(InhibitorNet.class));
	}

	// variant with ResetNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, ResetNet net, Marking state)
			throws ConnectionCannotBeObtained {
		return analyzeHomeMarkingPetriNet(context, net, state, PetrinetSemanticsFactory
				.regularResetNetSemantics(ResetNet.class));
	}

	// variant with ResetInhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, ResetInhibitorNet net, Marking state)
			throws ConnectionCannotBeObtained {
		return analyzeHomeMarkingPetriNet(context, net, state, PetrinetSemanticsFactory
				.regularResetInhibitorNetSemantics(ResetInhibitorNet.class));
	}

	/**
	 * Variant with net, marking, and semantics
	 */
	// variant with Petrinet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1, 3 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, Petrinet net, Marking state,
			PetrinetSemantics semantics) throws ConnectionCannotBeObtained {
		semantics.initialize(net.getTransitions(), new Marking(state));
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, null, semantics);
	}

	// variant with InhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1, 3 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, InhibitorNet net, Marking state,
			InhibitorNetSemantics semantics) throws ConnectionCannotBeObtained {
		semantics.initialize(net.getTransitions(), new Marking(state));
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, null, semantics);
	}

	// variant with ResetNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1, 3 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, ResetNet net, Marking state,
			ResetNetSemantics semantics) throws ConnectionCannotBeObtained {
		semantics.initialize(net.getTransitions(), new Marking(state));
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, null, semantics);
	}

	// variant with ResetInhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1, 3 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, ResetInhibitorNet net, Marking state,
			ResetInhibitorNetSemantics semantics) throws ConnectionCannotBeObtained {
		semantics.initialize(net.getTransitions(), new Marking(state));
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, null, semantics);
	}

	/**
	 * variants with net, marking, and reachability graph
	 */
	// variant with Petrinet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1, 2 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, Petrinet net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, reachabilityGraph);
	}

	// variant with InhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1, 2 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, InhibitorNet net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, reachabilityGraph);
	}

	// variant with ResetNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1, 2 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, ResetNet net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, reachabilityGraph);
	}

	// variant with ResetInhibitorNet and marking
	@PluginVariant(variantLabel = "Analyze Home Marking", requiredParameterLabels = { 0, 1, 2 })
	public HomeMarkingSet analyzeHomeMarkingPetriNet(PluginContext context, ResetInhibitorNet net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, reachabilityGraph);
	}

	/**
	 * Executed method when reachability graph is provided without semantics.
	 * Then, semantic is retrieved from connection
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be analyzed
	 * @param state
	 *            initial state (initial marking)
	 * @param reachabilityGraph
	 *            reachability graph of this net and marking
	 * @return HomeMarkingSet a set of home marking
	 * @throws Exception
	 */
	private HomeMarkingSet analyzeHomeMarkingPetriNetPrivate(PluginContext context, PetrinetGraph net, Marking state,
			ReachabilityGraph reachabilityGraph) throws ConnectionCannotBeObtained {
		// reachability graph must be produced by a certain semantics, therefore
		// try catch block is skipped
		StateSpaceConnection connection = context.getConnectionManager().getFirstConnection(StateSpaceConnection.class,
				context, net, state, reachabilityGraph);
		Semantics<Marking, Transition> sem = connection.getObjectWithRole(AbstractSemanticConnection.SEMANTICS);
		sem.initialize(net.getTransitions(), state);
		return analyzeHomeMarkingPetriNetPrivate(context, net, state, reachabilityGraph, sem);
	}

	/**
	 * Method to determine home marking for a net. Given net, marking, semantic,
	 * and reachability graph
	 * 
	 * @param context
	 *            context of the net
	 * @param net
	 *            net to be analyzed
	 * @param state
	 *            initial state (initial marking)
	 * @param reachabilityGraph
	 *            reachability graph of this net and marking
	 * @return HomeMarkingSet a set of home marking
	 * @throws Exception
	 */
	private HomeMarkingSet analyzeHomeMarkingPetriNetPrivate(PluginContext context, PetrinetGraph net, Marking state,
			ReachabilityGraph reachabilityGraph, Semantics<Marking, Transition> semantics)
			throws ConnectionCannotBeObtained {
		ConnectionManager cm = context.getConnectionManager();

		// check connectivity between marking and net
		cm.getFirstConnection(InitialMarkingConnection.class, context, net, state);

		// get reachability graph
		// in case no reachabilityGraph provided, means semantic is provided
		// check whether there is a corresponding reachability graph related to
		// the semantic
		if (reachabilityGraph == null) {
			try {
				reachabilityGraph = context.tryToFindOrConstructFirstObject(ReachabilityGraph.class,
						StateSpaceConnection.class, StateSpaceConnection.STATEPACE, net, state, semantics);
			} catch (ConnectionCannotBeObtained e) {
				// if the reachability graph cannot be constructed, home marking can not be detected
				HomeMarkingSet set = new HomeMarkingSet(new Marking[0]);
				context.log("Statespace construction to identify home marking failed. Maximum number of states reached");
				
				// add connection
				context.addConnection(new HomeMarkingConnection(net, state, set, semantics, "Home Markings"));
				context.getFutureResult(0).setLabel("Home Markings of " + net.getLabel());
				return set;
			} catch (Exception exc) {
				context.log("Exception happened. Maximum number of states reached");
			}
			
		}

		// variable to hold the result
		Set<Marking> homeMarkings = new HashSet<Marking>();

		// get shortest path
		ShortestPathInfo<State, org.processmining.models.graphbased.directed.transitionsystem.Transition> shortestPathInfo = ShortestPathFactory
				.calculateAllShortestDistanceDijkstra(reachabilityGraph);

		checkState: for (State stateSpace1 : reachabilityGraph.getNodes()) {
			// check in all mappings, is there any value goes to stateSpace
			for (State stateSpace2 : reachabilityGraph.getNodes()) {
				if (shortestPathInfo.getShortestPathLength(stateSpace2, stateSpace1) < 0) {
					continue checkState;
				}
			}
			homeMarkings.add((Marking) stateSpace1.getIdentifier());
		}

		// transform to array
		Marking[] marking = new Marking[homeMarkings.size()];
		int counter = 0;
		for (Marking m : homeMarkings) {
			marking[counter] = m;
			counter++;
		}
		HomeMarkingSet set = new HomeMarkingSet(marking);

		// add connection
		context.addConnection(new HomeMarkingConnection(net, state, set, semantics, "Home Markings"));
		context.getFutureResult(0).setLabel("Home Markings of " + net.getLabel());
		return set;
	}
}