package org.processmining.plugins.transitionsystem.miner.ui;

import info.clearthought.layout.TableLayout;
import info.clearthought.layout.TableLayoutConstants;

import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeSet;

import javax.swing.ButtonGroup;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.tree.DefaultMutableTreeNode;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.model.XLog;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.models.graphbased.AttributeMap;
import org.processmining.plugins.transitionsystem.converter.util.TSConversions;
import org.processmining.plugins.transitionsystem.miner.TSMiner;
import org.processmining.plugins.transitionsystem.miner.TSMinerInput;
import org.processmining.plugins.transitionsystem.miner.TSMinerOutput;
import org.processmining.plugins.transitionsystem.miner.TSMinerPlugin;
import org.processmining.plugins.transitionsystem.miner.modir.TSMinerModirInput;
import org.processmining.plugins.transitionsystem.miner.util.TSAbstractions;
import org.processmining.plugins.transitionsystem.miner.util.TSDirections;

import com.fluxicon.slickerbox.colors.SlickerColors;
import com.fluxicon.slickerbox.components.NiceSlider;
import com.fluxicon.slickerbox.components.NiceSlider.Orientation;
import com.fluxicon.slickerbox.factory.SlickerDecorator;
import com.fluxicon.slickerbox.factory.SlickerFactory;

public class TSMinerUI {
	private final UIPluginContext context;
	private TSMinerInput settings;
	private Collection<XEventClassifier> classifiers;
	private int introductionStep;
	private int identificationStep;
	private int abstractionStep;
	private int[] classifierStep; // There can be as many classifier steps as classifiers.
	private int transitionStep;
	private int attributeStep;
	private int conversionStep;
	private int overviewStep;
	private int nofSteps;
	private myStep[] mySteps;
	private int currentStep;

	private TSAbstractions abstraction;
	private int horizon;

	public TSMinerUI(final UIPluginContext context) {
		this.context = context;
	}

	public Object[] mine(XLog log, Collection<XEventClassifier> classifiers, XEventClassifier transitionClassifier) {
		TSMinerInput input = new TSMinerInput(context, log, classifiers, transitionClassifier);
		settings = input;
		this.classifiers = classifiers;
		TSMinerOutput output;
		InteractionResult result = InteractionResult.NEXT;

		nofSteps = 0;
		introductionStep = nofSteps++;
		identificationStep = nofSteps++;
		abstractionStep = nofSteps++;
		classifierStep = new int[classifiers.size()];
		for (int i = 0; i < classifiers.size(); i++) {
			classifierStep[i] = nofSteps++;
		}
		attributeStep = nofSteps++;
		transitionStep = nofSteps++;
		conversionStep = nofSteps++;
		overviewStep = nofSteps++;

		mySteps = new myStep[nofSteps];
		mySteps[introductionStep] = new IntroductionStep();
		mySteps[identificationStep] = new IdentificationStep();
		mySteps[abstractionStep] = new AbstractionStep();
		int r = 0;
		for (XEventClassifier classifier : classifiers) {
			mySteps[classifierStep[r++]] = new ClassifierStep(classifier);
		}
		mySteps[attributeStep] = new AttributeStep();
		mySteps[transitionStep] = new TransitionStep();
		mySteps[conversionStep] = new ConversionStep();
		mySteps[overviewStep] = new OverviewStep();

		currentStep = introductionStep;

		while (true) {
			if (currentStep < 0) {
				currentStep = 0;
			}
			if (currentStep >= nofSteps) {
				currentStep = nofSteps - 1;
			}
//			context.log("Current step: " + currentStep);
			result = context
					.showWizard("TS Miner", currentStep == 0, currentStep == nofSteps - 1, mySteps[currentStep]);
			switch (result) {
				case NEXT :
					mySteps[currentStep].readSettings();
					go(1);
					break;
				case PREV :
					go(-1);
					break;
				case FINISHED :
					TSMiner miner = new TSMiner(context);
					output = miner.mine(settings);
					TSMinerPlugin.setLabels(context, log);
					output.getTransitionSystem().getAttributeMap().put(AttributeMap.LABEL,
							context.getFutureResult(0).getLabel());
					return new Object[] { output.getTransitionSystem(), output.getWeights(), output.getStarts(),
							output.getAccepts() };
				default :
					context.getFutureResult(0).cancel(true);
					context.getFutureResult(1).cancel(true);
					context.getFutureResult(2).cancel(true);
					context.getFutureResult(3).cancel(true);
					return new Object[] { null, null, null, null };
			}
		}
	}

	private int go(int direction) {
		currentStep += direction;
		if ((currentStep >= 0) && (currentStep < nofSteps)) {
			if (mySteps[currentStep].precondition()) {
				return currentStep;
			} else {
				return go(direction);
			}
		}
		return currentStep;
	}

	private abstract class myStep extends JPanel {
		/**
		 * 
		 */
		private static final long serialVersionUID = 6892655601953727616L;

		public abstract boolean precondition();

		public abstract void readSettings();
	}

	private class IntroductionStep extends myStep {
		/**
		 * 
		 */
		private static final long serialVersionUID = 8620447135225476025L;

		public IntroductionStep() {
			initComponents();
		}

		public boolean precondition() {
			return true;
		}

		private void initComponents() {
			double size[][] = { { TableLayoutConstants.FILL }, { TableLayoutConstants.FILL } };
			setLayout(new TableLayout(size));
			String body = "<p>This wizard will guide you through the process of configuring this miner.</p>";
			body += "<p>The configuration options for this miner can be divided into three categories:<ol>";
			body += "<li>options for configuring state keys (states differ if and only if their keys differ),</li>";
			body += "<li>options for configuring transition labels, and</li>";
			body += "<li>options for configuring post-mining conversions.</li></ol>";
			body += "The wizard will allow you to configure these three categories in the given order.</p>";
			add(SlickerFactory.instance().createLabel("<html><h1>Introduction</h1>" + body), "0, 0, l, t");
		}

		public void readSettings() {
		}
	}

	private class IdentificationStep extends myStep {
		/**
		 * 
		 */
		private static final long serialVersionUID = -5629896729801647063L;

		public IdentificationStep() {
			initComponents();
		}

		public boolean precondition() {
			return true;
		}

		private void initComponents() {
			double size[][] = { { TableLayoutConstants.FILL, 20, TableLayoutConstants.FILL },
					{ 50, 30, TableLayoutConstants.FILL, 30 } };
			setLayout(new TableLayout(size));
			add(SlickerFactory.instance().createLabel("<html><h2>Configure key classifiers</h2>"), "0, 0, 2, 0");

			add(SlickerFactory.instance().createLabel("<html><h3>Select backward keys</h3>"), "0, 1");
			JScrollPane jScrollPane2 = new javax.swing.JScrollPane();
			SlickerDecorator.instance().decorate(jScrollPane2, SlickerColors.COLOR_BG_3, SlickerColors.COLOR_FG,
					SlickerColors.COLOR_BG_1);
			jList1 = new javax.swing.JList(classifiers.toArray());
			jList1.setSelectionInterval(0, classifiers.size() - 1);
			jScrollPane2.setPreferredSize(new Dimension(250, 300));
			jScrollPane2.setViewportView(jList1);
			add(jScrollPane2, "0, 2");

			add(SlickerFactory.instance().createLabel("<html><h3>Select forward keys</h3>"), "2, 1");
			JScrollPane jScrollPane3 = new javax.swing.JScrollPane();
			SlickerDecorator.instance().decorate(jScrollPane3, SlickerColors.COLOR_BG_3, SlickerColors.COLOR_FG,
					SlickerColors.COLOR_BG_1);
			jList2 = new javax.swing.JList(classifiers.toArray());
			jScrollPane3.setPreferredSize(new Dimension(250, 300));
			jScrollPane3.setViewportView(jList2);
			add(jScrollPane3, "2, 2");

			jCheckBox1 = SlickerFactory.instance().createCheckBox("Select key data attributes", false);
			add(jCheckBox1, "0, 3, 2, 3");
		}

		private javax.swing.JCheckBox jCheckBox1;
		private javax.swing.JList jList1;
		private javax.swing.JList jList2;

		public void readSettings() {
			int i = 0;
			for (XEventClassifier classifier : classifiers) {
				settings.getModirSettings(TSDirections.BACKWARD, classifier).setUse(jList1.isSelectedIndex(i));
				settings.getModirSettings(TSDirections.FORWARD, classifier).setUse(jList2.isSelectedIndex(i));
				i++;
			}
			settings.setUseAttributes(jCheckBox1.isSelected());
		}
	}

	private class AbstractionStep extends myStep implements ActionListener {
		/**
		 * 
		 */
		private static final long serialVersionUID = 6933657976761947420L;

		public AbstractionStep() {
			initComponents();
		}

		public boolean precondition() {
			for (XEventClassifier classifier : settings.getClassifiers()) {
				for (TSDirections direction : TSDirections.values()) {
					if (settings.getModirSettings(direction, classifier).getUse()) {
						return true;
					}
				}
			}
			return false;
		}

		private void initComponents() {
			double size[][] = { { 54, TableLayoutConstants.FILL },
					{ 50, 30, 20, 20, 20, 30, 20, 20, TableLayoutConstants.FILL } };
			setLayout(new TableLayout(size));
			add(SlickerFactory.instance().createLabel("<html><h2>Configure key classifier collections</h2>"),
					"0, 0, 1, 0");
			add(SlickerFactory.instance().createLabel("<html><h3>Select collection type</h2>"), "0, 1, 1, 1");

			useSequence = SlickerFactory.instance().createRadioButton("List");
			useMultiset = SlickerFactory.instance().createRadioButton("Multiset");
			useSet = SlickerFactory.instance().createRadioButton("Set");
			ButtonGroup typeGroup = new ButtonGroup();
			typeGroup.add(useSequence);
			typeGroup.add(useMultiset);
			typeGroup.add(useSet);
			add(useSequence, "0, 2, 1, 2");
			add(useMultiset, "0, 3, 1, 3");
			add(useSet, "0, 4, 1, 4");
			useSet.setSelected(true);

			add(SlickerFactory.instance().createLabel("<html><h3>Select collection size limit</h2>"), "0, 5, 1, 5");
			allEventsRadioButton = SlickerFactory.instance().createRadioButton("No limit");
			nofEventsRadioButton = SlickerFactory.instance().createRadioButton("Limit");
			nofEventsSlider = SlickerFactory.instance().createNiceIntegerSlider("", 0, 100, 1, Orientation.HORIZONTAL);
			nofEventsSlider.setPreferredSize(new Dimension(420, 20));
			ButtonGroup limitGroup = new ButtonGroup();
			limitGroup.add(allEventsRadioButton);
			limitGroup.add(nofEventsRadioButton);
			add(allEventsRadioButton, "0, 6, 1, 6");
			add(nofEventsRadioButton, "0, 7");
			add(nofEventsSlider, "1, 7");
			nofEventsRadioButton.setSelected(true);
			
			allEventsRadioButton.addActionListener(this);
			nofEventsRadioButton.addActionListener(this);
		}

		private javax.swing.JRadioButton allEventsRadioButton;
		private javax.swing.JRadioButton nofEventsRadioButton;
		private NiceSlider nofEventsSlider;
		private javax.swing.JRadioButton useMultiset;
		private javax.swing.JRadioButton useSequence;
		private javax.swing.JRadioButton useSet;

		public void readSettings() {
			TSMinerModirInput modirSettings;

			if (useSequence.isSelected()) {
				abstraction = TSAbstractions.SEQUENCE;
			} else if (useMultiset.isSelected()) {
				abstraction = TSAbstractions.BAG;
			} else {
				abstraction = TSAbstractions.SET;
			}
			if (allEventsRadioButton.isSelected()) {
				horizon = -1;
			} else {
				horizon = nofEventsSlider.getSlider().getValue();
			}
			/**
			 * Set abstraction and (visible) horizon for all selected
			 * combinations of modes (model element, originator, event type) and
			 * directions (backward, forward).
			 */
			for (TSDirections direction : TSDirections.values()) {
				for (XEventClassifier classifier : settings.getClassifiers()) {
					modirSettings = settings.getModirSettings(direction, classifier);
					if (modirSettings.getUse()) {
						modirSettings.setAbstraction(abstraction);
						modirSettings.setFilteredHorizon(horizon);
						modirSettings.setHorizon(-1);
					}
				}
			}
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			// TODO Auto-generated method stub
			nofEventsSlider.setEnabled(e.getSource() == nofEventsRadioButton);
		}
	}

	private class ClassifierStep extends myStep {

		/**
		 * 
		 */
		private static final long serialVersionUID = -7916606698553515246L;
		private final XEventClassifier classifier;
		Collection<String> s;
		Map<String, Integer> m;
		NiceSlider slider;
		
		public ClassifierStep(XEventClassifier classifier) {
			this.classifier = classifier;
			initComponents();
		}

		public boolean precondition() {
			return (settings.getModirSettings(TSDirections.BACKWARD, classifier).getUse() || settings.getModirSettings(
					TSDirections.FORWARD, classifier).getUse());
		}

		private void initComponents() {
			double size[][] = { { TableLayoutConstants.FILL }, { 50, 30, TableLayoutConstants.FILL , 10, 20} };
			setLayout(new TableLayout(size));
			add(SlickerFactory.instance().createLabel("<html><h2>Configure key classifier filter</h2>"), "0, 0");
			add(SlickerFactory.instance().createLabel("<html><h3>Select '" + classifier.name() + "' values</h3>"),
					"0, 1");

			jScrollPane1 = new javax.swing.JScrollPane();
			SlickerDecorator.instance().decorate(jScrollPane1, SlickerColors.COLOR_BG_3, SlickerColors.COLOR_FG,
					SlickerColors.COLOR_BG_1);
			XLogInfo info = settings.getLogInfo();
			s = new TreeSet<String>();
			m = new HashMap<String, Integer>();
			for (XEventClass eventClass : info.getEventClasses(classifier).getClasses()) {
				s.add(eventClass.toString());
				m.put(eventClass.toString(), eventClass.size());
			}
			nameFilter = new javax.swing.JList(s.toArray());
			nameFilter.setSelectionInterval(0, info.getEventClasses(classifier).getClasses().size() - 1);
			jScrollPane1.setPreferredSize(new Dimension(450, 300));
			jScrollPane1.setViewportView(nameFilter);

			add(jScrollPane1, "0, 2");

			slider = SlickerFactory.instance().createNiceIntegerSlider("Select top percentage", 0, 100, 80,
					Orientation.HORIZONTAL);
			ChangeListener listener = new ChangeListener() {

				public void stateChanged(ChangeEvent e) {
					// TODO Auto-generated method stub
					int percentage = slider.getSlider().getValue();
					int size = 0;
					TreeSet<Integer> eventSizes = new TreeSet<Integer>();
					for (String event : s) {
						size += m.get(event);
						eventSizes.add(m.get(event));
					}
					int treshold = size * percentage;
					int value = 0;
					nameFilter.clearSelection();
					while (100 * value < treshold) {
						int eventSize = eventSizes.last();
						eventSizes.remove(eventSize);
						int index = 0;
						for (String event : s) {
							if (m.get(event) == eventSize) {
								value += eventSize;
								nameFilter.addSelectionInterval(index, index);
							}
							index++;
						}
					}
				}

			};
			slider.addChangeListener(listener);
			listener.stateChanged(null);
			add(slider, "0, 4");
}

		private javax.swing.JScrollPane jScrollPane1;
		private javax.swing.JList nameFilter;

		public void readSettings() {
			for (TSDirections direction : TSDirections.values()) {
				if (settings.getModirSettings(direction, classifier).getUse()) {
					settings.getModirSettings(direction, classifier).setUse(!nameFilter.isSelectionEmpty());
				}
				settings.getModirSettings(direction, classifier).getFilter().clear();
				for (Object object : nameFilter.getSelectedValues()) {
					if (settings.getModirSettings(direction, classifier).getUse()) {
						settings.getModirSettings(direction, classifier).getFilter().add(object.toString());
					}
				}
			}
		}
	}

	private class TransitionStep extends myStep {
		/**
		 * 
		 */
		private static final long serialVersionUID = 5295381313201819465L;
		Collection<String> s;
		Map<String, Integer> m;
		NiceSlider slider;

		public TransitionStep() {
			initComponents();
		}

		public boolean precondition() {
			return true;
		}

		private void initComponents() {
			double size[][] = { { TableLayoutConstants.FILL }, { 50, 30, TableLayoutConstants.FILL, 10, 20 } };
			setLayout(new TableLayout(size));
			add(SlickerFactory.instance().createLabel("<html><h2>Configure transition label filter</h2>"), "0, 0");
			add(SlickerFactory.instance().createLabel("<html><h3>Select transition label values</h3>"), "0, 1");

			jScrollPane1 = new javax.swing.JScrollPane();
			XLogInfo info = settings.getLogInfo();
			s = new TreeSet<String>();
			m = new HashMap<String, Integer>();
			/*
			 * The transition classifier has been as as default classifier while
			 * creating the log info. As a result, getEventClasses() return the
			 * classes for this classifier.
			 */
			for (XEventClass eventClass : info.getEventClasses().getClasses()) {
				s.add(eventClass.toString());
				m.put(eventClass.toString(), eventClass.size());
			}
			nameFilter = new javax.swing.JList(s.toArray());
			nameFilter.setSelectionInterval(0, info.getEventClasses().getClasses().size() - 1);
			jScrollPane1.setPreferredSize(new Dimension(450, 300));
			jScrollPane1.setViewportView(nameFilter);
			add(jScrollPane1, "0, 2");

			slider = SlickerFactory.instance().createNiceIntegerSlider("Select top percentage", 0, 100, 80,
					Orientation.HORIZONTAL);
			ChangeListener listener = new ChangeListener() {

				public void stateChanged(ChangeEvent e) {
					// TODO Auto-generated method stub
					int percentage = slider.getSlider().getValue();
					int size = 0;
					TreeSet<Integer> eventSizes = new TreeSet<Integer>();
					for (String event : s) {
						size += m.get(event);
						eventSizes.add(m.get(event));
					}
					int treshold = size * percentage;
					int value = 0;
					nameFilter.clearSelection();
					while (100 * value < treshold) {
						int eventSize = eventSizes.last();
						eventSizes.remove(eventSize);
						int index = 0;
						for (String event : s) {
							if (m.get(event) == eventSize) {
								value += eventSize;
								nameFilter.addSelectionInterval(index, index);
							}
							index++;
						}
					}
				}

			};
			slider.addChangeListener(listener);
			listener.stateChanged(null);
			add(slider, "0, 4");
		}

		private javax.swing.JScrollPane jScrollPane1;
		private javax.swing.JList nameFilter;

		public void readSettings() {
			settings.getVisibleFilter().clear();
			for (Object object : nameFilter.getSelectedValues()) {
				settings.getVisibleFilter().add(object.toString());
			}
		}
	}

	private class AttributeStep extends myStep {
		/**
		 * 
		 */
		private static final long serialVersionUID = 7223716214843258396L;

		public AttributeStep() {
			initComponents();
		}

		public boolean precondition() {
			return (settings.getUseAttributes());
		}

		private void initComponents() {
			double size[][] = { { TableLayoutConstants.FILL }, { 50, 30, TableLayoutConstants.FILL } };
			setLayout(new TableLayout(size));
			add(SlickerFactory.instance().createLabel("<html><h2>Configure key data attributes</h2>"), "0, 0");
			add(SlickerFactory.instance().createLabel("<html><h3>Select key data attributes</h3>"), "0, 1");

			jScrollPane1 = new javax.swing.JScrollPane();
			attributeFilter = new javax.swing.JList(settings.getAttributeFilter().toArray());
			attributeFilter.setSelectionInterval(0, settings.getAttributeFilter().size() - 1);
			jScrollPane1.setPreferredSize(new Dimension(450, 300));
			jScrollPane1.setViewportView(attributeFilter);
			add(jScrollPane1, "0, 2");
		}

		private javax.swing.JScrollPane jScrollPane1;
		private javax.swing.JList attributeFilter;

		public void readSettings() {
			settings.setUseAttributes(!attributeFilter.isSelectionEmpty());
			settings.getAttributeFilter().clear();
			for (Object object : attributeFilter.getSelectedValues()) {
				settings.getAttributeFilter().add(object.toString());
			}
		}
	}

	private class ConversionStep extends myStep {
		/**
		 * 
		 */
		private static final long serialVersionUID = 4974107535166248951L;

		public ConversionStep() {
			initComponents();
		}

		public boolean precondition() {
			return true;
		}

		private void initComponents() {
			double size[][] = { { TableLayout.FILL }, { 50, 20, 20, 20, 20, TableLayoutConstants.FILL } };
			setLayout(new TableLayout(size));
			add(SlickerFactory.instance().createLabel("<html><h2>Configure post-mining conversions</h2>"), "0, 0");

			removeSelfLoopsCheckBox = new javax.swing.JCheckBox();
			improveDiamondCheckBox = new javax.swing.JCheckBox();
			mergeInflowCheckBox = new javax.swing.JCheckBox();
			mergeOutflowCheckBox = new javax.swing.JCheckBox();

			removeSelfLoopsCheckBox = SlickerFactory.instance().createCheckBox("Remove self loops", true);
			improveDiamondCheckBox = SlickerFactory.instance().createCheckBox("Improve diamond structure", false);
			mergeInflowCheckBox = SlickerFactory.instance().createCheckBox("Merge states with identical inflow", false);
			mergeOutflowCheckBox = SlickerFactory.instance().createCheckBox("Merge states with identical outflow",
					false);
			add(removeSelfLoopsCheckBox, "0, 1");
			add(improveDiamondCheckBox, "0, 2");
			add(mergeInflowCheckBox, "0, 3");
			add(mergeOutflowCheckBox, "0, 4");

		}

		private javax.swing.JCheckBox improveDiamondCheckBox;
		private javax.swing.JCheckBox mergeInflowCheckBox;
		private javax.swing.JCheckBox mergeOutflowCheckBox;
		private javax.swing.JCheckBox removeSelfLoopsCheckBox;

		public void readSettings() {
			settings.getConverterSettings().setUse(TSConversions.KILLSELFLOOPS, removeSelfLoopsCheckBox.isSelected());
			settings.getConverterSettings().setUse(TSConversions.EXTEND, improveDiamondCheckBox.isSelected());
			settings.getConverterSettings().setUse(TSConversions.MERGEBYINPUT, mergeInflowCheckBox.isSelected());
			settings.getConverterSettings().setUse(TSConversions.MERGEBYOUTPUT, mergeOutflowCheckBox.isSelected());
		}
	}

	private class OverviewStep extends myStep {
		/**
		 * 
		 */
		private static final long serialVersionUID = -2853400854055327898L;

		public OverviewStep() {
			initComponents();
		}

		public boolean precondition() {
			makingVisible();
			return true;
		}

		private void initComponents() {
			double size[][] = { { TableLayout.FILL }, { 50, TableLayout.FILL } };
			setLayout(new TableLayout(size));
			add(SlickerFactory.instance().createLabel("<html><h2>Check configuration</h2>"), "0, 0");

			jScrollPane1 = new javax.swing.JScrollPane();
			jScrollPane1.setPreferredSize(new Dimension(450, 350));
			jScrollPane1.setViewportView(new javax.swing.JTree());
			add(jScrollPane1, "0, 1");
		}

		private javax.swing.JScrollPane jScrollPane1;

		/**
		 * Updates the overview just before it is being displayed.
		 */
		public void makingVisible() {
			/**
			 * Construct a new tree.
			 */
			DefaultMutableTreeNode topNode = new DefaultMutableTreeNode("TS Miner configuration");
			DefaultMutableTreeNode idNode = new DefaultMutableTreeNode("Key classifiers");
			topNode.add(idNode);
			for (XEventClassifier classifier : classifiers) {
				if (settings.getModirSettings(TSDirections.BACKWARD, classifier).getUse()) {
					DefaultMutableTreeNode nfNode = new DefaultMutableTreeNode(classifier + " backwards");
					idNode.add(nfNode);
					for (String name : settings.getModirSettings(TSDirections.BACKWARD, classifier).getFilter()) {
						nfNode.add(new DefaultMutableTreeNode(name));
					}
				}
				if (settings.getModirSettings(TSDirections.FORWARD, classifier).getUse()) {
					DefaultMutableTreeNode nfNode = new DefaultMutableTreeNode(classifier + " forwards");
					idNode.add(nfNode);
					for (String name : settings.getModirSettings(TSDirections.FORWARD, classifier).getFilter()) {
						nfNode.add(new DefaultMutableTreeNode(name));
					}
				}
			}
			DefaultMutableTreeNode abNode = new DefaultMutableTreeNode("Collection type");
			idNode.add(abNode);
			if (abstraction == TSAbstractions.SEQUENCE) {
				abNode.add(new DefaultMutableTreeNode("List"));
			} else if (abstraction == TSAbstractions.BAG) {
				abNode.add(new DefaultMutableTreeNode("Multiset"));
			} else {
				abNode.add(new DefaultMutableTreeNode("Set"));
			}
			abNode = new DefaultMutableTreeNode("Collection size");
			idNode.add(abNode);
			if (horizon == -1) {
				abNode.add(new DefaultMutableTreeNode("No limit"));
			} else {
				abNode.add(new DefaultMutableTreeNode("Limit: " + horizon));
			}
			if (settings.getUseAttributes()) {
				DefaultMutableTreeNode nfNode = new DefaultMutableTreeNode("Key data attributes");
				idNode.add(nfNode);
				for (String name : settings.getAttributeFilter()) {
					nfNode.add(new DefaultMutableTreeNode(name));
				}
			}
			DefaultMutableTreeNode tNode = new DefaultMutableTreeNode("Transition label filter");
			topNode.add(tNode);
			for (String name : settings.getVisibleFilter()) {
				tNode.add(new DefaultMutableTreeNode(name));
			}
			DefaultMutableTreeNode coNode = new DefaultMutableTreeNode("Post-mining conversions");
			topNode.add(coNode);
			boolean b = false;
			if (settings.getConverterSettings().getUse(TSConversions.KILLSELFLOOPS)) {
				coNode.add(new DefaultMutableTreeNode("Remove self loops"));
				b = true;
			}
			if (settings.getConverterSettings().getUse(TSConversions.EXTEND)) {
				coNode.add(new DefaultMutableTreeNode("Improve diamond structure"));
				b = true;
			}
			if (settings.getConverterSettings().getUse(TSConversions.MERGEBYINPUT)) {
				coNode.add(new DefaultMutableTreeNode("Merge states with identical inflow"));
				b = true;
			}
			if (settings.getConverterSettings().getUse(TSConversions.MERGEBYOUTPUT)) {
				coNode.add(new DefaultMutableTreeNode("Merge states with identical outflow"));
				b = true;
			}
			if (!b) {
				coNode.add(new DefaultMutableTreeNode("None"));
			}
			/**
			 * Display the new tree.
			 */
			jScrollPane1.setViewportView(new javax.swing.JTree(topNode));
		}

		public void readSettings() {
		}
	}
}
