package org.processmining.processtree;

import java.util.Collection;
import java.util.List;
import java.util.UUID;

/**
 * Nodes have a name and an ID. They are uniquely identified through their ID,
 * not through their name.
 * 
 * @author bfvdonge
 * 
 */
public interface Node extends ProcessTreeElement {

	/**
	 * returns an editable collection of variables that are read by this node
	 * 
	 * @return
	 */
	public Collection<Variable> getReadVariables();
	
	/**
	 * adds the read variable to this node
	 * @param var
	 * @return
	 */
	public boolean addReadVariable(Variable var);
	
	/**
	 * removes the read variable from this node
	 * @param var
	 * @return
	 */
	public boolean removeReadVariable(Variable var);
	
	/**
	 * returns an editable collection of variables which can be removed from the list of read variables
	 * @return
	 */
	public Collection<Variable> getRemovableReadVariables();
	
	
	/**
	 * adds the removable read variable to this node
	 * @param var
	 * @return
	 */
	public boolean addRemovableReadVariable(Variable var);
	
	/**
	 * removes the removable read variable to this node
	 * @param var
	 * @return
	 */
	public boolean removeRemovableReadVariable(Variable var);
	
	/**
	 * returns the collection of variables read in this subgraph
	 * 
	 * @return
	 */
	public Collection<Variable> getReadVariablesRecursive();
	
	/**
	 * returns an editable collection of variables that are written by this node
	 * 
	 * @return
	 */
	public Collection<Variable> getWrittenVariables();
	
	/**
	 * adds the write variable to this node
	 * @param var
	 * @return
	 */
	public boolean addWriteVariable(Variable var);
	
	/**
	 * removes the write variable from this node
	 * @param var
	 * @return
	 */
	public boolean removeWriteVariable(Variable var);
	
	/**
	 * returns an editable collection of variables which can be removed from the list of written variables
	 * @return
	 */
	public Collection<Variable> getRemovableWrittenVariables();
	
	/**
	 * adds the removable write variable to this node
	 * @param var
	 * @return
	 */
	public boolean addRemovableWriteVariable(Variable var);
	
	/**
	 * removes the removable write variable to this node
	 * @param var
	 * @return
	 */
	public boolean removeRemovableWriteVariable(Variable var);
	
	/**
	 * returns the collection of variables written in this subgraph
	 * 
	 * @return
	 */
	public Collection<Variable> getWrittenVariablesRecursive();

	/**
	 * sets the processtree to which this node belongs
	 * 
	 * @param tree
	 */
	public void setProcessTree(ProcessTree tree);

	/**
	 * returns the process tree to which this node belongs
	 * 
	 * @return
	 */
	public ProcessTree getProcessTree();

	/**
	 * adds a new parent with given id and given expression to this node as the rightmost parent
	 * 
	 * @param id
	 * @param parent
	 * @param expression
	 * @return
	 */
	public Edge addParent(UUID id, Block parent, Expression expression);

	/**
	 * adds a new parent with random id and given expression to this node as the rightmost parent
	 * 
	 * @param parent
	 * @param expression
	 * @return
	 */
	public Edge addParent(Block parent, Expression expression);

	/**
	 * adds a new parent with random id and empty expression to this node as the rightmost parent
	 * 
	 * @param parent
	 * @return
	 */
	public Edge addParent(Block parent);

	/**
	 * returns an collection of blocks that are the parents of this node in the
	 * tree
	 * 
	 * changes to the collection are NOT reflected in the tree
	 * 
	 * @return
	 */
	public Collection<Block> getParents();
	
	/**
	 * removes an incoming edge from this node
	 * 
	 * @return
	 */
	public boolean removeIncomingEdge(Edge edge);

	/**
	 * returns the number of parents
	 */
	public int numParents();

	/**
	 * returns an immutable list of incoming edges.
	 * 
	 * @return
	 */
	public List<Edge> getIncomingEdges();

	/**
	 * Adds an incoming edge to the node
	 * 
	 * @param edge
	 */
	public void addIncomingEdge(Edge edge);

	/**
	 * returns true if this node has no parents.
	 * 
	 * @return
	 */
	public boolean isRoot();

	/**
	 * returns a short string representation of the subclass
	 * 
	 * @return
	 */
	public String toStringShort();

	/**
	 * returns true if this node is a leaf node.
	 * 
	 * @return
	 */
	public boolean isLeaf();
	
}