package org.processmining.processtree.configuration.controlflow.impl;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import org.processmining.framework.util.Pair;
import org.processmining.processtree.Block;
import org.processmining.processtree.Block.PlaceHolder;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.configuration.ConfigurationAuxiliary;
import org.processmining.processtree.configuration.controlflow.ControlFlowConfiguration;
import org.processmining.processtree.configuration.controlflow.ControlFlowConfigurationConstraints;
import org.processmining.processtree.configuration.controlflow.ControlFlowConfigurationIterator;
import org.processmining.processtree.impl.ProcessTreeImpl;

public class ControlFlowConfigurationIteratorImpl implements ControlFlowConfigurationIterator{

	private ProcessTree tree;
	public BitSet currentConfiguration;
	public ArrayList<UUID> confOpsH;
	public ArrayList<UUID> confOpsB;
	public ArrayList<Pair<UUID, UUID>> replOps;
	public BitSet lastReturnedConfiguration = null;
	private int bitLength;
	private BitSet minConfiguration = null;
	private BitSet maxConfiguration = null;
	private boolean doubleOption;
	private ControlFlowConfigurationConstraints constraints = null;
	
	/*
	 * Things to consider:
	 *  + Start with placeholders (easier to subdivide the work, else we need to know the first bit of the last placeholder node)
	 *  + Order the Vectors -> ArrayList
	 *  + Sort the UUID's
	 *  + Make a constructor with start and end bitsequence
	 *  - Add support to return intervals of bitsequences based on a given subdivision -> we do not really want this
	 */

	public ControlFlowConfigurationIteratorImpl(ProcessTree tree) throws Exception{
		this(tree, new ControlFlowConfigurationConstraintsImpl(), null, null);
	}
	
	public ControlFlowConfigurationIteratorImpl(ProcessTree tree, ControlFlowConfigurationConstraints constraints, Object minConfiguration, Object maxConfiguration){
		this.tree = new ProcessTreeImpl(tree);
		this.constraints = constraints;
		this.minConfiguration = (BitSet) minConfiguration;
		this.maxConfiguration = (BitSet) maxConfiguration;
		fillBlockedHidden();		
		fillRelacementOptions();
		this.bitLength = getBitLength();
		sort();
		currentConfiguration = new BitSet(bitLength + 2); // last 2 bits is for denoting termination, and fixing the size
		currentConfiguration.clear();
		currentConfiguration.set(bitLength + 1, true);
		if(this.minConfiguration != null){
			currentConfiguration = (BitSet) this.minConfiguration.clone();
		}
	}
	
	private void fillBlockedHidden(){
		confOpsH = new ArrayList<UUID>();
		confOpsB = new ArrayList<UUID>();
		for(Edge e: tree.getEdges()){
			if(e.isHideable()){
				confOpsH.add(e.getID());
			}
			if(e.isBlockable()){
				confOpsB.add(e.getID());
			}
		}
	}
	
	private void fillRelacementOptions(){
		replOps = new ArrayList<Pair<UUID, UUID>>();
		for(Node n: tree.getNodes()){
			if(n instanceof PlaceHolder){
				for(Edge e: ((Block)n).getOutgoingEdges()){
					replOps.add(new Pair<UUID, UUID>(n.getID(), e.getTarget().getID()));
				}
			}
		}
	}
	
	private int getBitLength(){
		int bitLength = confOpsH.size() + confOpsB.size();
		Set<PlaceHolder> placeHolderNodes = new HashSet<PlaceHolder>();
		HashMap<PlaceHolder, Integer> nrOfOptions = new HashMap<PlaceHolder, Integer>();
		for(Node n: tree.getNodes()){
			if(n instanceof PlaceHolder){
				placeHolderNodes.add((PlaceHolder) n);
			}
		}
		for(PlaceHolder p: placeHolderNodes){
			nrOfOptions.put(p, p.getOutgoingEdges().size());
		}
		for(PlaceHolder p: placeHolderNodes){
			bitLength += nrOfOptions.get(p);
		}
		return bitLength;
	}
	
	private void sort(){
		Collections.sort(confOpsH);
		Collections.sort(confOpsB);
		// now sort the replacement options
		ArrayList<UUID> placeholderNodes = new ArrayList<UUID>();
		HashMap<UUID, List<UUID>> replOptions = new HashMap<UUID, List<UUID>>();
		for(Pair<UUID, UUID> repl: replOps){
			if(!placeholderNodes.contains(repl.getFirst())){
				placeholderNodes.add(repl.getFirst());
			}
			List<UUID> options = replOptions.get(repl.getFirst());
			if(options == null){
				options = new ArrayList<UUID>();
			}
			options.add(repl.getSecond());
			replOptions.put(repl.getFirst(), options);
		}
		// now we have the structure we want
		replOps.clear();
		Collections.sort(placeholderNodes);
		for(UUID id: placeholderNodes){
			List<UUID> ids = replOptions.get(id);
			Collections.sort(ids);
			// now we have sorted everything, lets add them
			for(UUID repId: ids){
				replOps.add(new Pair<UUID, UUID>(id, repId));
			}
		}
	}
	
	public ControlFlowConfiguration nextConfiguration() throws Exception{
		ControlFlowConfiguration conf = new ControlFlowConfigurationImpl();
		if(lastReturnedConfiguration == null){
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		if(currentConfiguration.equals(lastReturnedConfiguration) && !currentConfiguration.get(currentConfiguration.length() - 2)){
			ConfigurationAuxiliary.nextBitSequece(currentConfiguration);
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		while((!isConsistentComplete(conf, tree) && !currentConfiguration.get(currentConfiguration.length() - 2) && ConfigurationAuxiliary.smaller(currentConfiguration, maxConfiguration))){
			ConfigurationAuxiliary.nextBitSequece(currentConfiguration);
			conf = new ControlFlowConfigurationImpl();
			fillConfiguration(conf);
			makeComplete(conf, tree);
		}
		if(isConsistentComplete(conf, tree) && !currentConfiguration.get(currentConfiguration.length() - 2) && !(currentConfiguration.equals(lastReturnedConfiguration)) && ConfigurationAuxiliary.smaller(currentConfiguration, maxConfiguration)){
			lastReturnedConfiguration = (BitSet) currentConfiguration.clone();
			return conf;
		}
		else{
			return null;
		}
	}
	
	public boolean hasNext() throws Exception{
		BitSet currentC = (BitSet) currentConfiguration.clone();
		BitSet lastRetConf = null;
		if(lastReturnedConfiguration != null){
			lastRetConf = (BitSet) lastReturnedConfiguration.clone();
		}
		if(nextConfiguration() != null){
			currentConfiguration = currentC;
			lastReturnedConfiguration = lastRetConf;
			return true;
		}
		else{
			return false;
		}
	}
	
	public ControlFlowConfiguration getNext() throws Exception {
		if(hasNext()){
			return nextConfiguration();
		}
		throw new NullPointerException();
	}
	
	public void reset(){
		currentConfiguration = new BitSet(bitLength + 2); // last 2 bits is for denoting termination, and fixing the size
		currentConfiguration.clear();
		currentConfiguration.set(bitLength + 1, true);
		if(this.minConfiguration != null){
			currentConfiguration = (BitSet) this.minConfiguration.clone();
		}
		lastReturnedConfiguration = null;
	}

	private void fillConfiguration(ControlFlowConfiguration conf){
		// note the -2 due to the control bits
		// we need to keep track of another boolean because else we do not know if a replacement option is set twice
		doubleOption = false;
		HashMap<UUID, Boolean> optionSet = new HashMap<UUID, Boolean>();
		for(int i = 0; i < currentConfiguration.length() - 2; i++){
			if(currentConfiguration.get(i)){
				if(i < replOps.size()){
					Pair<UUID, UUID> currentRel = replOps.get(i);
					conf.setSelectedReplacement(currentRel.getFirst(), currentRel.getSecond());
					if(optionSet.get(currentRel.getFirst()) != null){
						doubleOption = true;
					}
					optionSet.put(currentRel.getFirst(), true);
				}
				else if(i < confOpsH.size() + replOps.size() && replOps.size() <= i){
					conf.addHiddenEdge(confOpsH.get(i - replOps.size()));
				}
				else{
					conf.addBlockedEdge(confOpsB.get(i - (replOps.size() + confOpsH.size())));
				}
			}
		}
	}
			
	protected static void makeComplete(ControlFlowConfiguration conf, ProcessTree tree){
		for(Edge e: tree.getEdges()){
			if(e.isHideable()){
				if(!conf.getHiddenEdges().contains(e.getID())){
					conf.addNonHiddenEdge(e);
				}
			}
			if(e.isBlockable()){
				if(!conf.getBlockedEdges().contains(e.getID())){
					conf.addNonBlockedEdge(e);
				}
			}
		}
		for(Node n: tree.getNodes()){
			if(n instanceof PlaceHolder){
				if(((PlaceHolder) n).getOutgoingEdges().size() == 1){
					// we have a very special case, initialise an empty array in the nonSelectedReplacementOptions
					conf.getNonSelectedReplacements().put(n.getID(), new ArrayList<UUID>());
				}
				for(Edge e: ((PlaceHolder)n).getOutgoingEdges()){
					if(conf.getSelectedReplacements().get(n.getID()) == null){
						// just stop
						break;
					}
					if(!conf.getSelectedReplacements().get(n.getID()).equals(e.getTarget().getID())){
						conf.addNonSelectedReplacement((PlaceHolder)n, e.getTarget());
					}
				}
			}
		}
	}
	
	private boolean isConsistentComplete(ControlFlowConfiguration conf, ProcessTree tree){
		if(ConfigurationAuxiliary.overlap(conf.getBlockedEdges(), conf.getNonBlockedEdges())){
			return false;
		}
		if(ConfigurationAuxiliary.overlap(conf.getHiddenEdges(), conf.getNonHiddenEdges())){
			return false;
		}
		if(ConfigurationAuxiliary.overlap(conf.getHiddenEdges(), conf.getBlockedEdges())){
			return false;
		}
		for(UUID p: conf.getSelectedReplacements().keySet()){
			if(conf.getNonSelectedReplacements().get(p).contains(conf.getSelectedReplacements().get(p))){
				return false;
			}
		}
		for(Node n: tree.getNodes()){
			if(n instanceof PlaceHolder){
				if(conf.getSelectedReplacements().get(n.getID()) == null){
					return false;
				}
				if(conf.getNonSelectedReplacements().get(n.getID()).size() != (((Block)n).getOutgoingEdges().size() - 1)){
					return false;
				}
			}
		}
		// another check because we no longer have a list of replacements, i.e., the last replacement option is done twice (hence extra boolean)
		if(doubleOption){
			return false;
		}
		return constraints.isValid(conf);
		//return true;
	}
	
	public List<Pair<Object, Object>> getSubdivisions(int divisions) throws Exception {
		// we have some bit fun here :-D
		//List<Pair<Object, Object>> subdivisions = new ArrayList<Pair<Object,Object>>();
		//BitSet min;
		//BitSet max = new BitSet();
		// easy, we get this one for free :D
		//min = (BitSet) currentConfiguration.clone();
		throw new Exception("Not implemented\n");
		//return null;
	}

}
