package org.processmining.processtree.configuration.impl;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import org.processmining.framework.util.Pair;
import org.processmining.plugins.properties.auxiliary.TransitivePreset;
import org.processmining.plugins.properties.processmodel.Property;
import org.processmining.plugins.properties.processmodel.configurable.ConfigurableProperty;
import org.processmining.plugins.properties.processmodel.configurable.ConfigurablePropertyConfiguration;
import org.processmining.plugins.properties.processmodel.configurable.ConfigurablePropertyIterator;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTreeElement;
import org.processmining.processtree.configuration.Configuration;
import org.processmining.processtree.configuration.ConfigurationIterator;
import org.processmining.processtree.configuration.controlflow.ControlFlowConfiguration;
import org.processmining.processtree.configuration.controlflow.ControlFlowConfigurationInstantiater;
import org.processmining.processtree.configuration.controlflow.ControlFlowConfigurationIterator;
import org.processmining.processtree.configuration.controlflow.impl.ControlFlowConfigurationInstantiaterImpl;
import org.processmining.processtree.configuration.controlflow.impl.ControlFlowConfigurationIteratorImpl;
import org.processmining.processtree.configuration.data.DataConfiguration;
import org.processmining.processtree.configuration.data.DataConfigurationInstantiater;
import org.processmining.processtree.configuration.data.DataConfigurationIterator;
import org.processmining.processtree.configuration.data.impl.DataConfigurationInstantiaterImpl;
import org.processmining.processtree.configuration.data.impl.DataConfigurationIteratorImpl;
import org.processmining.processtree.configuration.originator.OriginatorConfiguration;
import org.processmining.processtree.configuration.originator.OriginatorConfigurationInstantiater;
import org.processmining.processtree.configuration.originator.OriginatorConfigurationIterator;
import org.processmining.processtree.configuration.originator.impl.OriginatorConfigurationInstantiaterImpl;
import org.processmining.processtree.configuration.originator.impl.OriginatorConfigurationIteratorImpl;
import org.processmining.processtree.impl.ProcessTreeImpl;

/**
 * @author DSchunse
 * Does include properties internally, but these are not (yet) offered to the outside, this means the same configuration 
 * can be returned multiple times.
 */
public class ConfigurationIteratorImpl implements ConfigurationIterator{

	// we have an iterator over the different iterators of everything configurable
	// we have a predetermined set of iterators (cf, data, resources) and over properties
	
	// now think of something smart to iterate over this -> found it
	// we have a bitset 0 denotes still busy, 1 denotes reset and increment the first one with 0
	
	// we have non-local operations, i.e., by modifying the control-flow, we have to do everything again
	
	private ProcessTree treeOriginal;
	private ProcessTree tree;
	@Deprecated
	private BitSet currentConfiguration; // only for properties
	private int bitLength;
	private ControlFlowConfigurationIterator cfIt;
	private ControlFlowConfigurationInstantiater cfIn;
	private ControlFlowConfiguration cfC;
	private DataConfigurationIterator dIt;
	private DataConfigurationInstantiater dIn;
	private DataConfiguration dC;
	private OriginatorConfigurationIterator oIt;
	private OriginatorConfigurationInstantiater oIn;
	private OriginatorConfiguration oC;
	private List<UUID> elements;
	private Map<UUID, List<ConfigurableProperty<?>>> propsPerElement;
	private Map<UUID, Map<ConfigurableProperty<?>, ConfigurablePropertyIterator<?>>> epIt;
	private Map<UUID, Map<ConfigurableProperty<?>, ConfigurablePropertyConfiguration<?>>> epC;
	
	public ConfigurationIteratorImpl(ProcessTree tree) throws Exception{
		this.treeOriginal = tree;
		this.tree = new ProcessTreeImpl(this.treeOriginal);
		this.cfIt = new ControlFlowConfigurationIteratorImpl(this.tree);
		this.cfIn = new ControlFlowConfigurationInstantiaterImpl();
		// a call to make it possible to iterate over the bitsets
		initialiseProperties();
	}
	
	private void initialiseProperties() throws Exception{
		this.epIt = new HashMap<UUID, Map<ConfigurableProperty<?>, ConfigurablePropertyIterator<?>>>();
		this.epC = new HashMap<UUID, Map<ConfigurableProperty<?>, ConfigurablePropertyConfiguration<?>>>();
		fillStructures();
		sortAndInitialise();
		instantiatePropertyIterators();
		this.bitLength = getBitLength();
		// now sort the different things
		currentConfiguration = new BitSet(bitLength + 2); // last 2 bits is for denoting termination, and fixing the size
		currentConfiguration.clear();
		currentConfiguration.set(bitLength + 1, true);
	}
	
	private void fillStructures() throws IllegalAccessException, InstantiationException{
		Set<UUID> elems = new HashSet<UUID>();
		for(ProcessTreeElement elem: tree.getElements()){
			Map<ConfigurableProperty<?>, ConfigurablePropertyIterator<?>> epItMap = epIt.get(elem);
			Map<ConfigurableProperty<?>, ConfigurablePropertyConfiguration<?>> epCMap = epC.get(elem);
			for(Property<?> prop: elem.getIndependentProperties().keySet()){
				if(prop instanceof ConfigurableProperty<?>){
					elems.add(elem.getID());
					if(epItMap == null){
						epItMap = new HashMap<ConfigurableProperty<?>, ConfigurablePropertyIterator<?>>();
					}
					if(epCMap == null){
						epCMap = new HashMap<ConfigurableProperty<?>, ConfigurablePropertyConfiguration<?>>();
					}
					epItMap.put((ConfigurableProperty<?>)prop, ((ConfigurableProperty<?>)prop).getIterator(elem.getIndependentProperty(prop)));
				}
				
			}
			if(epItMap != null){
				epIt.put(elem.getID(), epItMap);
			}
			if(epCMap != null){
				epC.put(elem.getID(), epCMap);
			}
		}
		elements = new ArrayList<UUID>(elems);
	}
	
	private void instantiatePropertyIterators() throws Exception{
		// we need to instantiate all iterators and apply these on the tree, note that we now really want the hierarchy of properties
		for(UUID elem: elements){
			for(ConfigurableProperty<?> cProp: propsPerElement.get(elem)){
				ConfigurablePropertyIterator<?> cpI = epIt.get(elem).get(cProp);
				ConfigurablePropertyConfiguration<?> cPropC = cpI.getNext();
				// we have the first instantiation, now update the tree
				ProcessTreeElement element = tree.getElement(elem);
				Object values = element.getIndependentProperty(cProp);
				element.removeIndependentProperty(cProp);
				element.setIndependentProperty(cProp.getNonConfigurableVariant().newInstance(), cProp.instantiateTotal(values, cPropC));
				epC.get(elem).put(cProp, cPropC);
			}
		}
	}
	
	private int getBitLength(){
		int bitLength = 0;
		for(UUID elem: elements){
			bitLength += epIt.get(elem).keySet().size();
		}
		return bitLength;
	}
	
	private void sortAndInitialise(){
		Collections.sort(elements);
		if(propsPerElement == null){
			propsPerElement = new HashMap<UUID, List<ConfigurableProperty<?>>>();
		}
		// we cannot sort the properties, so just make a list out of them
		for(UUID elem: elements){
			List<ConfigurableProperty<?>> confProps = TransitivePreset.computeOrderingProperties(epIt.get(elem).keySet());
			propsPerElement.put(elem, confProps);
		}
	}
	
	public boolean hasNext() throws Exception{
		return cfIt.hasNext() || dIt.hasNext() || oIt.hasNext() || hasNextProperty();
	}
	
	private boolean hasNextProperty() throws Exception{
		// this one is relatively straightforward, i.e., just go over all iterators and see if anyone still has something left
		// has reimplemented for a different purpose, so we can reuse that functionality
		return (determineCarryPosition() != -1);
	}
	/**
	 * 
	 * @return the next instance. Note that this can be NULL if we have an empty tree.
	 * @throws Exception
	 */
	public Configuration getNext() throws Exception{
		if(hasNext()){
			return nextConfiguration();
		}
		else{
			throw new NullPointerException();
		}
	}
	
	public Map<UUID, Set<UUID>> getNextMapping(){
		return cfIn.getMappingHiddenNodesToTaus();
	}
	
	public Configuration nextConfiguration() throws Exception{
		// so we have some initialisation, i.e., initially we only have a control-flow iterator
		int carryPosition = determineCarryPosition();
		if(dIt == null){
			// okay lets initialise the whole lot
			cfC = cfIt.getNext();
			tree = cfIn.instantiate(treeOriginal, cfC, true);
			// we have a special case when the tree is null
			if(tree == null){
				return null;
			}
			this.dIt = new DataConfigurationIteratorImpl(tree);		
			dC = dIt.getNext();
			this.dIn = new DataConfigurationInstantiaterImpl();
			tree = dIn.instantiate(tree, dC);
			
			this.oIt = new OriginatorConfigurationIteratorImpl(tree);
			oC = oIt.getNext();
			this.oIn = new OriginatorConfigurationInstantiaterImpl();
			tree = oIn.instantiate(tree, oC);
			initialiseProperties();
			Configuration conf = new ConfigurationImpl();
			conf.setControlFlowConfiguration(cfC);
			conf.setDataConfiguration(dC);
			conf.setOriginatorConfiguration(oC);
			return conf;
		}		
		else if(carryPosition == -1){
			// we are through everything, reset the iterators of the properties, and increment the Originators, Data, and Control-Flow
			// note that we might need to reset some iterators
			boolean resetP = true;
			boolean resetO = false;
			boolean resetD = false;
			boolean resetCF = false;
			resetO = resetP && !oIt.hasNext();
			resetD = resetO && !dIt.hasNext();
			resetCF = resetD && !cfIt.hasNext();
			if(resetCF){
				// we cannot have this, throw new null pointer exception
				throw new Exception("This cannot happen");
			}
			else if(resetD){
				// so we need to reinitialise D en O
				cfC = cfIt.getNext();
				tree = cfIn.instantiate(treeOriginal, cfC, true);
				// we have a special case when the tree is null
				if(tree == null){
					return null;
				}
				this.dIt = new DataConfigurationIteratorImpl(tree);		
				dC = dIt.getNext();
				tree = dIn.instantiate(tree, dC);
				
				this.oIt = new OriginatorConfigurationIteratorImpl(tree);
				oC = oIt.getNext();
				tree = oIn.instantiate(tree, oC);
				initialiseProperties();
				Configuration conf = new ConfigurationImpl();
				conf.setControlFlowConfiguration(cfC);
				conf.setDataConfiguration(dC);
				conf.setOriginatorConfiguration(oC);
				return conf;
			}
			else if(resetO){
				// we need to reinitialise O
				tree = cfIn.instantiate(treeOriginal, cfC, true);	
				dC = dIt.getNext();
				tree = dIn.instantiate(tree, dC);
				
				this.oIt = new OriginatorConfigurationIteratorImpl(tree);
				oC = oIt.getNext();
				tree = oIn.instantiate(tree, oC);
				initialiseProperties();
				Configuration conf = new ConfigurationImpl();
				conf.setControlFlowConfiguration(cfC);
				conf.setDataConfiguration(dC);
				conf.setOriginatorConfiguration(oC);
				return conf;
			}
			else if(resetP){
				// we need to reinitialise the properties
				tree = cfIn.instantiate(treeOriginal, cfC, true);
				tree = dIn.instantiate(tree, dC);
				oC = oIt.getNext();
				this.oIn = new OriginatorConfigurationInstantiaterImpl();
				tree = oIn.instantiate(tree, oC);
				initialiseProperties();
				Configuration conf = new ConfigurationImpl();
				conf.setControlFlowConfiguration(cfC);
				conf.setDataConfiguration(dC);
				conf.setOriginatorConfiguration(oC);
				return conf;
			}	
		}
		else{
			// NOTE THAT we do NOT generate a new configuration here, just fiddling with the properties
			
			// we need to apply an incremental approach on the tree
			tree = cfIn.instantiate(treeOriginal, cfC, true);		
			tree = dIn.instantiate(tree, dC);
			tree = oIn.instantiate(tree, oC);
		
			// so we have our list of bits, we can just start at the beginning
			// note we do not need the notion of a current configuration, i.e., the iterators determine the state
			int index = 0;
			for(UUID id: elements){
				for(ConfigurableProperty<?> confP: propsPerElement.get(id)){
					if(index < carryPosition){
						// do nothing, just initialise the tree	(done at the end)				
					}
					if(index > carryPosition){
						// reinitialise the iterator and update the configuration accordingly
						ProcessTreeElement elem = tree.getElement(id);
						epIt.get(id).put(confP, ((ConfigurableProperty<?>)confP).getIterator(elem.getIndependentProperty(confP)));
						
					}
					if(index >= carryPosition){
						// get the next value (note we are reusing some functionality
						ConfigurablePropertyIterator<?> confIt = epIt.get(id).get(confP);
						ConfigurablePropertyConfiguration<?> confC = confIt.getNext();
						epC.get(id).put(confP, confC);
					}
					ConfigurablePropertyConfiguration<?> cPropC = epC.get(id).get(confP);
					ProcessTreeElement element = tree.getElement(id);
					Object values = element.getIndependentProperty(confP);
					element.removeIndependentProperty(confP);
					element.setIndependentProperty(confP.getNonConfigurableVariant().newInstance(), confP.instantiateTotal(values, cPropC));
					index++;
				}
			}
			Configuration conf = new ConfigurationImpl();
			conf.setControlFlowConfiguration(cfC);
			conf.setDataConfiguration(dC);
			conf.setOriginatorConfiguration(oC);
			return conf;
		}
		throw new Exception("If we get here, the compiler is wrong!");
	}
	
	/**
	 * 
	 * @return the position of the last iterator which can be used, if -1 is returned we have gone through all of the iterators 
	 * @throws Exception
	 */
	private int determineCarryPosition() throws Exception{
		// TODO: this must be able to do it more efficient, i.e., the currentConfiguration has been made deprecated
		// we need to determine where the carry position will be
		// note the -3 due to control bits
		int index = currentConfiguration.length() - 3;
		// we have to start from the last element
		for(int i = elements.size() - 1; i >= 0; i--){
			UUID id = elements.get(i);
			for(int j = propsPerElement.get(id).size() - 1; j >= 0; j--){
				// guess we only need to look at the hasNext(), if hasNext() = false, we have a carry,
				// and then we'll need to progress until the carry becomes false
				ConfigurablePropertyIterator<?> confIt = epIt.get(id).get(propsPerElement.get(id).get(j));
				// we have a next for this index, so we can just return the index
				if(confIt.hasNext()){
					return index;
				}
				index--;
			}
		}
		// we did not find any position, we need to return -1, to signal we are through all of the iterators
		return -1;
	}

	public List<Pair<Object, Object>> getSubdivisions(int divisions) throws Exception {
		throw new Exception("Not implemented\n");
	}
}
