package org.processmining.processtree.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;

import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.Variable;

/**
 * implements a block which provides an iterator over its child nodes. The nodes
 * are returned by the iterator in the order they are in the tree.
 * 
 * @author bfvdonge
 * 
 */
public abstract class AbstractBlock extends AbstractNode implements Block {

	protected final List<Edge> outgoing;
	protected boolean changeable;

	public AbstractBlock(String name) {
		this(name, new ArrayList<Edge>(), new ArrayList<Edge>());
	}
	
	public AbstractBlock(UUID id, String name) {
		this(id, name, new ArrayList<Edge>(), new ArrayList<Edge>());
	}

	public AbstractBlock(String name, List<Edge> incoming, List<Edge> outgoing) {
		//super(name, incoming);
		//this.outgoing = outgoing;
		this(UUID.randomUUID(), name, incoming, outgoing);
	}
	
	public AbstractBlock(UUID id, String name, List<Edge> incoming, List<Edge> outgoing){
		super(id, name, incoming);
		this.outgoing = outgoing;
		this.changeable = false;
	}
	
	public AbstractBlock(Block b){
		super(b);
		this.outgoing = new ArrayList<Edge>();
		this.changeable = b.isChangeable();
	}	
	
	public Collection<Variable> getReadVariablesRecursive(){
		Collection<Variable> vars = new HashSet<Variable>();
		vars.addAll(readVars);
		for(Node n: getChildren()){
			vars.addAll(n.getReadVariablesRecursive());
		}
		return vars;
	}
	
	public Collection<Variable> getWrittenVariablesRecursive() {
		Collection<Variable> vars = new HashSet<Variable>();
		vars.addAll(writeVars);
		for(Node n: getChildren()){
			vars.addAll(n.getWrittenVariablesRecursive());
		}
		return vars;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Block#getOutgoingEdges()
	 */
	public List<Edge> getOutgoingEdges() {
		return Collections.unmodifiableList(outgoing);
	}

	public void addOutgoingEdge(Edge edge) {
		if(!outgoing.contains(edge)){
			outgoing.add(edge);
		}
	}
	
	public void removeOutgoingEdge(Edge edge){
		outgoing.remove(edge);
	}

	public void addOutgoingEdgeAt(Edge edge, int index) {
		if(outgoing.contains(edge)){
			outgoing.remove(edge);
		}
		outgoing.add(index, edge);
	}

	public boolean isLeaf() {
		return false;
	}
	
	public boolean isChangeable(){
		return changeable;
	}
	
	public void setChangeable(boolean changeable){
		this.changeable = changeable;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.processtree.Block#addChild(org.processmining.processtree
	 * .Node, org.processmining.processtree.Expression)
	 */
	public Edge addChild(Node child, Expression expression) {
		Edge e = new EdgeImpl(this, child, expression);
		tree.addEdge(e);
		addOutgoingEdge(e);
		readVars.addAll(expression.getVariables());
		return e;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.processtree.Block#addChild(org.processmining.processtree
	 * .Node)
	 */
	public Edge addChild(Node child) {
		return addChild(child, EdgeImpl.NOEXPRESSION);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.processtree.Block#addChildAt(org.processmining.processtree
	 * .Node, org.processmining.processtree.Expression, int)
	 */
	public Edge addChildAt(Node child, Expression expression, int index) {
		Edge e = new EdgeImpl(this, child, expression);
		tree.addEdge(e);
		addOutgoingEdgeAt(e, index);
		return e;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.processtree.Block#addChildAt(org.processmining.processtree
	 * .Node, int)
	 */
	public Edge addChildAt(Node child, int index) {
		return addChildAt(child, EdgeImpl.NOEXPRESSION, index);
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.processtree.Block#swapChildAt(org.processmining.processtree
	 * .Node, org.processmining.processtree.Expression, int)
	 */
	public Edge swapChildAt(Node child, Expression expression, int index) {
		Edge e = getOutgoingEdges().get(index);
		e.setTarget(child);
		return e;
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.processtree.Block#swapChildAt(org.processmining.processtree
	 * .Node, int)
	 */
	public Edge swapChildAt(Node child, int index) {
		return swapChildAt(child, EdgeImpl.NOEXPRESSION, index);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Block#getChildren()
	 */
	public List<Node> getChildren() {
		Node[] children = new Node[outgoing.size()];
		for (int i = outgoing.size(); i-- > 0;) {
			children[i] = outgoing.get(i).getTarget();
		}
		return Arrays.asList(children);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Block#iterator()
	 */
	public Iterator<Node> iterator() {
		return new Iterator<Node>() {
			private final Iterator<Edge> it = outgoing.iterator();

			@Override
			public boolean hasNext() {
				return it.hasNext();
			}

			@Override
			public Node next() {
				return it.next().getTarget();
			}

			@Override
			public void remove() {
				throw new UnsupportedOperationException("Cannot remove children through the iterator");
			}
		};
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.processmining.processtree.Block#numChildren()
	 */
	public int numChildren() {
		return outgoing.size();
	}
	
	public String toString() {
		return tree.toString(this);
	}

	/**
	 * Denotes an Xor block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static class Xor extends AbstractBlock implements Block.Xor {

		public Xor(String name) {
			super(name);
		}
		
		public Xor(UUID id, String name) {
			super(id, name);
		}

		public Xor(String name, List<Edge> incoming, List<Edge> outgoing) {
			super(name, incoming, outgoing);
		}
		
		public Xor(UUID id, String name, List<Edge> incoming, List<Edge> outgoing) {
			super(id, name, incoming, outgoing);
		}
		
		public Xor(Block.Xor b){
			super(b);
		}

		@Override
		public boolean orderingOfChildernMatters() {
			return true;
		}

		@Override
		public boolean expressionsOfOutgoingEdgesMatter() {
			return true;
		}

		@Override
		public String toStringShort() {
			return "Xor";
		}

	}

	/**
	 * Denotes an Or block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static class Or extends AbstractBlock implements Block.Or {
		public Or(String name) {
			super(name);
		}
		
		public Or(UUID id, String name) {
			super(id, name);
		}

		public Or(String name, List<Edge> incoming, List<Edge> outgoing) {
			super(name, incoming, outgoing);
		}
		
		public Or(UUID id, String name, List<Edge> incoming, List<Edge> outgoing) {
			super(id, name, incoming, outgoing);
		}
		
		public Or(Block.Or b){
			super(b);
		}

		@Override
		public boolean orderingOfChildernMatters() {
			return true;
		}

		@Override
		public boolean expressionsOfOutgoingEdgesMatter() {
			return true;
		}

		@Override
		public String toStringShort() {
			return "Or";
		}

	}

	/**
	 * Denotes an And block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static class And extends AbstractBlock implements Block.And {
		public And(String name) {
			super(name);
		}
		
		public And(UUID id, String name) {
			super(id, name);
		}

		public And(String name, List<Edge> incoming, List<Edge> outgoing) {
			super(name, incoming, outgoing);
		}
		
		public And(UUID id, String name, List<Edge> incoming, List<Edge> outgoing) {
			super(id, name, incoming, outgoing);
		}
		
		public And(Block.And b){
			super(b);
		}

		@Override
		public boolean orderingOfChildernMatters() {
			return false;
		}

		@Override
		public boolean expressionsOfOutgoingEdgesMatter() {
			return false;
		}

		@Override
		public String toStringShort() {
			return "And";
		}

	}

	/**
	 * Denotes a sequence block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static class Seq extends AbstractBlock implements Block.Seq {
		public Seq(String name) {
			super(name);
		}

		public Seq(UUID id, String name) {
			super(id, name);
		}

		public Seq(String name, List<Edge> incoming, List<Edge> outgoing) {
			super(name, incoming, outgoing);
		}
		
		public Seq(UUID id, String name, List<Edge> incoming, List<Edge> outgoing) {
			super(id, name, incoming, outgoing);
		}
		
		public Seq(Block.Seq b){
			super(b);
		}

		@Override
		public boolean orderingOfChildernMatters() {
			return true;
		}

		@Override
		public boolean expressionsOfOutgoingEdgesMatter() {
			return false;
		}

		@Override
		public String toStringShort() {
			return "Seq";
		}

	}

	/**
	 * Denotes a Xor Loop block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static class XorLoop extends AbstractBlock implements Block.XorLoop {
		public XorLoop(String name) {
			super(name);
		}

		public XorLoop(UUID id, String name) {
			super(id, name);
		}

		public XorLoop(String name, List<Edge> incoming, List<Edge> outgoing) {
			super(name, incoming, outgoing);
		}
		
		public XorLoop(UUID id, String name, List<Edge> incoming, List<Edge> outgoing) {
			super(id, name, incoming, outgoing);
		}
		
		public XorLoop(Block.XorLoop b){
			super(b);
		}

		@Override
		public boolean orderingOfChildernMatters() {
			return true;
		}

		@Override
		public boolean expressionsOfOutgoingEdgesMatter() {
			return true;
		}

		@Override
		public String toStringShort() {
			return "XorLoop";
		}

	}

	/**
	 * Denotes an deferred choice block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static class Def extends AbstractBlock implements Block.Def {
		public Def(String name) {
			super(name);
		}

		public Def(UUID id, String name) {
			super(id, name);
		}

		public Def(String name, List<Edge> incoming, List<Edge> outgoing) {
			super(name, incoming, outgoing);
		}
		
		public Def(UUID id, String name, List<Edge> incoming, List<Edge> outgoing) {
			super(id, name, incoming, outgoing);
		}
		
		public Def(Block.Def b){
			super(b);
		}

		@Override
		public boolean orderingOfChildernMatters() {
			return true;
		}

		@Override
		public boolean expressionsOfOutgoingEdgesMatter() {
			return false;
		}

		@Override
		public String toStringShort() {
			return "Def";
		}

	}

	/**
	 * Denotes an deferred choice loop block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static class DefLoop extends AbstractBlock implements Block.DefLoop {
		public DefLoop(String name) {
			super(name);
		}

		public DefLoop(UUID id, String name) {
			super(id, name);
		}

		public DefLoop(String name, List<Edge> incoming, List<Edge> outgoing) {
			super(name, incoming, outgoing);
		}
		
		public DefLoop(UUID id, String name, List<Edge> incoming, List<Edge> outgoing) {
			super(id, name, incoming, outgoing);
		}
		
		public DefLoop(Block.DefLoop b){
			super(b);
		}

		@Override
		public boolean orderingOfChildernMatters() {
			return true;
		}

		@Override
		public boolean expressionsOfOutgoingEdgesMatter() {
			return false;
		}

		@Override
		public String toStringShort() {
			return "DefLoop";
		}

	}

	/**
	 * Denotes an PlaceHolder block
	 * 
	 * @author bfvdonge
	 * 
	 */
	public static class PlaceHolder extends AbstractBlock implements Block.PlaceHolder {

		public PlaceHolder(String name) {
			super(name);
		}

		public PlaceHolder(UUID id, String name) {
			super(id, name);
		}

		public PlaceHolder(String name, List<Edge> incoming, List<Edge> outgoing) {
			super(name, incoming, outgoing);
		}
		
		public PlaceHolder(UUID id, String name, List<Edge> incoming, List<Edge> outgoing) {
			super(id, name, incoming, outgoing);
		}
		
		public PlaceHolder(Block.PlaceHolder b){
			super(b);
		}

		@Override
		public boolean orderingOfChildernMatters() {
			return false;
		}

		@Override
		public boolean expressionsOfOutgoingEdgesMatter() {
			return true;
		}

		@Override
		public String toStringShort() {
			return "P.H.";
		}

	}

}
