package org.processmining.processtree.ptml;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;

import org.deckfour.xes.extension.XExtension;
import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.extension.std.XOrganizationalExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.deckfour.xes.model.impl.XAttributeMapImpl;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.Originator;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Variable;
import org.xmlpull.v1.XmlPullParser;

/**
 * Basic (E)PTML object. Allows import, export
 * 
 * @author hverbeek
 * 
 */
public class Ptml extends PtmlElement {

	/**
	 * PTML tag.
	 */
	public final static String TAG = "ptml";

	private List<PtmlProcessTree> treeList;

	private XLog log;
	private XTrace trace;
	private XFactory factory;
	private XExtension conceptExtension;
	private XExtension organizationalExtension;

	boolean hasErrors;

	public Ptml() {
		super(TAG);
		treeList = new ArrayList<PtmlProcessTree>();

		initializeLog();
	}
	
	public UUID getId(){
		if(treeList.isEmpty()){
			return null;
		}
		else{
			return treeList.get(0).id;
		}
	}
	
	public String getName(){
		if(treeList.isEmpty()){
			return "";
		}
		else{
			return treeList.get(0).name;
		}
	}

	private void initializeLog() {
		factory = XFactoryRegistry.instance().currentDefault();
		conceptExtension = XConceptExtension.instance();
		organizationalExtension = XOrganizationalExtension.instance();
		log = factory.createLog();
		log.getExtensions().add(conceptExtension);
		log.getExtensions().add(organizationalExtension);

		log("<preamble>");

		hasErrors = false;
	}

	public XLog getLog() {
		return log;
	}

	public void log(String context, int lineNumber, String message) {
		XAttributeMap attributeMap = new XAttributeMapImpl();
		attributeMap.put(XConceptExtension.KEY_NAME,
				factory.createAttributeLiteral(XConceptExtension.KEY_NAME, message, conceptExtension));
		attributeMap.put(XConceptExtension.KEY_INSTANCE,
				factory.createAttributeLiteral(XConceptExtension.KEY_INSTANCE, context, conceptExtension));
		attributeMap.put(XOrganizationalExtension.KEY_RESOURCE, factory.createAttributeLiteral(
				XOrganizationalExtension.KEY_RESOURCE, "Line " + lineNumber, organizationalExtension));
		XEvent event = factory.createEvent(attributeMap);
		trace.add(event);
		hasErrors = true;
	}

	public void log(String name) {
		trace = factory.createTrace();
		log.add(trace);
		trace.getAttributes().put(XConceptExtension.KEY_NAME,
				factory.createAttributeLiteral(XConceptExtension.KEY_NAME, name, conceptExtension));
	}

	public boolean hasErrors() {
		return hasErrors;
	}

	protected boolean importElements(PluginContext context, XmlPullParser xpp, Ptml ptml) {
		if (super.importElements(context, xpp, ptml)) {
			return true;
		}
		if (xpp.getName().equals(PtmlProcessTree.TAG)) {
			PtmlProcessTree tree = new PtmlProcessTree(null);
			tree.importElement(context, xpp, ptml);
			treeList.add(tree);
			return true;
		}
		return false;
	}

	protected String exportElements(Ptml ptml) {
		String s = super.exportElements(ptml);
		for (PtmlProcessTree tree : treeList) {
			s += tree.exportElement(ptml);
		}
		return s;
	}
	
	public void unmarshall(ProcessTree tree) {
		for (PtmlProcessTree t : treeList) {
			t.unmarshall(tree, new HashMap<UUID, Node>(), new HashMap<UUID, Edge>(), new HashMap<UUID, Originator>(),
					new HashMap<UUID, Variable>(), new HashMap<UUID, Expression>());
		}
	}

	public Ptml marshall(List<ProcessTree> trees) {
		treeList = new ArrayList<PtmlProcessTree>();
		for (ProcessTree tree : trees) {
			PtmlProcessTree t = new PtmlProcessTree(tree.getID());
			t.marshall(tree);
			treeList.add(t);
		}
		return this;
	}

	public Ptml marshall(ProcessTree tree) {
		treeList = new ArrayList<PtmlProcessTree>();
		PtmlProcessTree t = new PtmlProcessTree(tree.getID());
		t.marshall(tree);
		treeList.add(t);
		return this;
	}
	
	public void checkAcyclicity() {
		for (PtmlProcessTree tree : treeList) {
			tree.checkAcyclicity(this);
		}
	}
}
