package org.processmining.processtree.visualization.bpmn;

import gnu.trove.list.TIntList;
import gnu.trove.list.array.TIntArrayList;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import javax.swing.BorderFactory;
import javax.swing.Icon;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.SwingConstants;

import org.deckfour.xes.classification.XEventClasses;
import org.jgraph.JGraph;
import org.jgraph.graph.DefaultEdge;
import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultGraphModel;
import org.jgraph.graph.DefaultPort;
import org.jgraph.graph.Edge.Routing;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.GraphLayoutCache;
import org.jgraph.graph.GraphModel;
import org.processmining.processtree.Block;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Event;
import org.processmining.processtree.Expression;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.ProcessTree.Type;
import org.processmining.processtree.Task;
import org.processmining.processtree.Task.Automatic;
import org.processmining.processtree.test.TestProcessTree;
import org.processmining.processtree.visualization.Dim;
import org.processmining.processtree.visualization.Path;
import org.processmining.processtree.visualization.SeSeGroup;
import org.processmining.processtree.visualization.extrainforenderer.ExtraInfoRenderer;
import org.processmining.processtree.visualization.jgraph.BPMNOperatorRenderer.OperatorType;
import org.processmining.processtree.visualization.jgraph.EventRenderer.EventType;
import org.processmining.processtree.visualization.jgraph.TreeCellViewFactory;

import com.jgraph.components.labels.CellConstants;
import com.jgraph.components.labels.MultiLineVertexRenderer;

public class TreeBPMNLayoutBuilder {

	@SuppressWarnings("unused")
	private static final long serialVersionUID = 1L;
	private final ProcessTree tree;
	private final Map<Path, Dim> dimensions;
	protected List<ExtraInfoRenderer> renderers;

	public static enum EdgeRouting {
		HORIZONTAL_FIRST, VERTICAL_FIRST, NORMAL, SKIP;
	}

	public static final int PERFORMANCEWIDTH = 100;
	public static final int PERFORMANCEHEIGHT = 50;
	
	public static final int BORDER = 12;
	public static final int LEAFHEIGHT = 50 + 2 * BORDER;
	public int LEAFWIDTH = 80 + 2 * BORDER;
	public int OPERATORHEIGHT = 40 + 2 * BORDER;
	public int OPERATORWIDTH = 40 + 2 * BORDER;
	
	public static final int ARROWSIZE = 7;
	public static final int CONFIGSIZE = 3 * ARROWSIZE;

	public static final int EVENTWIDTH = 30 + 2 * BORDER;
	public static final int EVENTHEIGHT = 30 + 2 * BORDER;
	private static final Routing ROUTING = new OrthogonalRouting();

	public static final double SPLINEPOINTFRACTION = .5;
	private final static int STYLE = GraphConstants.STYLE_ORTHOGONAL;
	public static final String EDGEROUTING = "EdgeRouting";
	public static final String BLOCKABLE = "Blockable";
	public static final String HIDEABLE = "Hideable";
	public static final String ROUTESKIPY = "skip edge y";
	@SuppressWarnings("unused")
	private final XEventClasses classes;

	//	private final static int STYLE = GraphConstants.STYLE_BEZIER;
	//	private final static int STYLE =  GraphConstants.STYLE_SPLINE;

	public TreeBPMNLayoutBuilder(ProcessTree tree) {
		this(tree, null);
	}

	public TreeBPMNLayoutBuilder(ProcessTree tree, XEventClasses classes) {
		this.tree = tree;
		this.classes = classes;
		this.dimensions = new HashMap<Path, Dim>(tree.getEdges().size() + 1);
		this.renderers = new ArrayList<ExtraInfoRenderer>();
		buildLayout();
	}
	
	public TreeBPMNLayoutBuilder(ProcessTree tree, List<ExtraInfoRenderer> renderers, XEventClasses classes) {
		this.tree = tree;
		this.renderers = renderers;
		this.classes = classes;
		this.dimensions = new HashMap<Path, Dim>(tree.getEdges().size() + 1);
		if(renderers == null){
			this.renderers = new ArrayList<ExtraInfoRenderer>();
		}
		if(renderers.size() > 0){
			OPERATORHEIGHT = PERFORMANCEHEIGHT;
			OPERATORWIDTH = PERFORMANCEWIDTH;
			LEAFWIDTH = Math.max(80 + 2 * BORDER, PERFORMANCEWIDTH + 2 * BORDER);
		}
		buildLayout();
	}

	public void buildLayout() {
		Stack<Edge> e = new Stack<Edge>();
		getSizeOfNode(e, true);
		setLocation(e, EVENTWIDTH, 0);
		//System.out.println("Tree: " + tree);
		//System.out.println("Tree dimensions: " + dimensions);
	}

	protected Dim getSizeOfNode(Stack<Edge> edges, boolean left2right) {
		Node node;
		if (edges.isEmpty()) {
			node = tree.getRoot();
		} else {
			node = edges.peek().getTarget();
		}
		// special case of the editor 
		if(node == null){
			Dim dim =  new Dim(0, 0, left2right);
			dimensions.put(null, dim);
			return dim;
		}
		Path key = new Path(edges);
		Dim dim;
		switch (tree.getType(node)) {
			case LOOPXOR :
			case LOOPDEF :
				dim = getSizeLoop(edges, (Block) node, left2right);
				break;
			case SEQ :
				dim = getSizeSeq(edges, (Block) node, left2right);
				break;
			case AND :
			case OR :
			case DEF :
			case XOR :
			case PLACEHOLDER :
				dim = getSizeOther(edges, (Block) node, left2right);
				break;
			case MESSAGE :
			case TIMEOUT :
				dim = getSizeEvent(edges, (Event) node, left2right);
				break;
			case AUTOTASK :
			case MANTASK :
			default :
				dim = getSizeLeaf(edges, (Task) node, left2right);
		}
		if (!edges.isEmpty() && (edges.peek().isBlockable() || edges.peek().isHideable())) {
			dim.width += CONFIGSIZE;
		}

		dimensions.put(key, dim);

		if (edges.isEmpty()) {
			dim = new Dim(dimensions.get(edges).width + 2 * EVENTWIDTH, dimensions.get(edges).height, dimensions.get(
					edges).isLeft2right());
			dimensions.put(null, dim);
		}
		return dim;
	}

	protected Dim getSizeSeq(Stack<Edge> edges, final Block seqNode, final boolean left2right) {
		int width = 0;
		int height = 0;

		Dim d;
		for (Edge e : seqNode.getOutgoingEdges()) {
			edges.push(e);
			d = getSizeOfNode(edges, left2right);
			edges.pop();
			width += d.width;
			if (d.height > height) {
				height = d.height;
			}
		}
		return new Dim(width, height, left2right);
	}

	protected Dim getSizeOther(Stack<Edge> edges, final Block otherNode, final boolean left2right) {
		int width = 0;
		int height = 0;

		Dim d;
		for (Edge e : otherNode.getOutgoingEdges()) {
			edges.push(e);
			d = getSizeOfNode(edges, left2right);
			edges.pop();
			height += d.height;
			if (d.width > width) {
				width = d.width;
			}
		}
		return new Dim(width + 2 * OPERATORWIDTH, height, left2right);
	}

	protected Dim getSizeLeaf(Stack<Edge> edges, final Task taskNode, final boolean left2right) {
		if (taskNode instanceof Task.Automatic) {
			Automatic auto = (Automatic) taskNode;
			if (auto.getReadVariables().isEmpty() && auto.getWrittenVariables().isEmpty()) {
				return new Dim(0, 0, left2right);
			}
		}
		return new Dim(LEAFWIDTH, LEAFHEIGHT, left2right);
	}

	protected Dim getSizeEvent(Stack<Edge> edges, final Event eventNode, final boolean left2right) {
		// size of children
		edges.push(eventNode.getOutgoingEdges().get(0));
		Dim d = getSizeOfNode(edges, left2right);
		edges.pop();

		return new Dim(d.width + EVENTWIDTH + BORDER, Math.max(EVENTHEIGHT, d.height), left2right);
	}

	protected Dim getSizeLoop(Stack<Edge> edges, final Block loopNode, final boolean left2right) {
		int width = 2 * OPERATORWIDTH;
		int height = 0;

		edges.push(loopNode.getOutgoingEdges().get(0));
		Dim leftDim = getSizeOfNode(edges, left2right);
		edges.pop();
		edges.push(loopNode.getOutgoingEdges().get(1));
		Dim middleDim = getSizeOfNode(edges, !left2right);
		edges.pop();
		edges.push(loopNode.getOutgoingEdges().get(2));
		Dim rightDim = getSizeOfNode(edges, left2right);
		edges.pop();

		width += rightDim.width;
		width += Math.max(leftDim.width, middleDim.width);

		height = Math.max(rightDim.height, middleDim.height + leftDim.height);

		return new Dim(width, height, left2right);
	}

	protected void setLocation(final Stack<Edge> edges, int x, int y) {
		// set the location of the target node of this edge
		Node node;
		if (edges.isEmpty()) {
			node = tree.getRoot();
		} else {
			node = edges.peek().getTarget();
		}
		// special case for the editor
		if(node == null){
			return;
		}
		Dim dim = dimensions.get(edges);
		dim.setLocation(x, y);

		if (!edges.isEmpty() && (edges.peek().isBlockable() || edges.peek().isHideable())) {
			x += CONFIGSIZE;
			dim.width -= CONFIGSIZE;
		}

		switch (tree.getType(node)) {
			case LOOPXOR :
			case LOOPDEF :
				setLocationLoop(edges, (Block) node, x, y);
				break;
			case SEQ :
				setLocationSeq(edges, (Block) node, x, y);
				break;
			case AND :
			case OR :
			case XOR :
			case DEF :
			case PLACEHOLDER :
				setLocationOther(edges, (Block) node, x, y);
				break;
			case MESSAGE :
			case TIMEOUT :
				setLocationEvent(edges, (Event) node, x, y);
				break;
			case AUTOTASK :
			case MANTASK :
			default :
				setLocationLeaf(edges, (Task) node, x, y);
		}
		if (!edges.isEmpty() && (edges.peek().isBlockable() || edges.peek().isHideable())) {
			dim.width += CONFIGSIZE;
		}
	}

	protected void setLocationLeaf(Stack<Edge> edges, final Task leafNode, final int x, final int y) {
		// skip;
	}

	protected void setLocationEvent(Stack<Edge> edges, final Event eventNode, final int x, final int y) {
		Dim dim = dimensions.get(edges);
		edges.push(eventNode.getOutgoingEdges().get(0));
		if (dim.isLeft2right()) {
			setLocation(edges, x + EVENTWIDTH + BORDER, y);
		} else {
			setLocation(edges, x, y);
		}
		edges.pop();
	}

	protected void setLocationLoop(Stack<Edge> edges, final Block loopNode, final int x, final int y) {
		// set the location for this node's children
		Edge left = loopNode.getOutgoingEdges().get(0);
		Edge middle = loopNode.getOutgoingEdges().get(1);
		Edge right = loopNode.getOutgoingEdges().get(2);
		edges.push(left);
		List<Edge> leftIndex = new ArrayList<Edge>(edges);
		edges.pop();
		edges.push(middle);
		List<Edge> middleIndex = new ArrayList<Edge>(edges);
		edges.pop();
		edges.push(right);
		List<Edge> rightIndex = new ArrayList<Edge>(edges);
		edges.pop();

		int h1, h2;
		if (dimensions.get(rightIndex).height < dimensions.get(leftIndex).height + dimensions.get(middleIndex).height) {
			h1 = 0;
			h2 = (dimensions.get(leftIndex).height + dimensions.get(middleIndex).height - dimensions.get(rightIndex).height) / 2;
		} else {
			h2 = 0;
			h1 = (dimensions.get(rightIndex).height - dimensions.get(leftIndex).height - dimensions.get(middleIndex).height) / 2;
		}

		if (dimensions.get(edges).isLeft2right()) {
			if (dimensions.get(leftIndex).width < dimensions.get(middleIndex).width) {
				edges.push(left);
				setLocation(edges, //
						x + OPERATORWIDTH + (dimensions.get(middleIndex).width - dimensions.get(leftIndex).width) / 2,// 
						y + h1);
				edges.pop();
				edges.push(middle);
				setLocation(edges, //
						x + OPERATORWIDTH,//
						y + h1 + dimensions.get(leftIndex).height);
				edges.pop();
				edges.push(right);
				setLocation(edges, //
						x + 2 * OPERATORWIDTH + dimensions.get(middleIndex).width, //
						y + h2);
				edges.pop();
			} else {
				edges.push(left);
				setLocation(edges, //
						x + OPERATORWIDTH, // 
						y + h1);
				edges.pop();
				edges.push(middle);
				setLocation(edges,//
						x + OPERATORWIDTH + (dimensions.get(leftIndex).width - dimensions.get(middleIndex).width) / 2,//
						y + dimensions.get(leftIndex).height + h1);
				edges.pop();
				edges.push(right);
				setLocation(edges, //
						x + 2 * OPERATORWIDTH + dimensions.get(leftIndex).width, //
						y + h2);
				edges.pop();
			}
		} else {
			int wx = OPERATORWIDTH + dimensions.get(rightIndex).width;
			if (dimensions.get(leftIndex).width < dimensions.get(middleIndex).width) {
				edges.push(left);
				setLocation(edges, //
						x + wx + (dimensions.get(middleIndex).width - dimensions.get(leftIndex).width) / 2,// 
						y + h1);
				edges.pop();
				edges.push(middle);
				setLocation(edges,//
						x + wx,//
						y + dimensions.get(leftIndex).height + h1);
				edges.pop();
				edges.push(right);
				setLocation(edges, //
						x, //
						y + h2);
				edges.pop();
			} else {
				edges.push(left);
				setLocation(edges, //
						x + wx, // 
						y + h1);
				edges.pop();
				edges.push(middle);
				setLocation(edges,//
						x + wx + (dimensions.get(leftIndex).width - dimensions.get(middleIndex).width) / 2,//
						y + dimensions.get(leftIndex).height + h1);
				edges.pop();
				edges.push(right);
				setLocation(edges, //
						x, //
						y + h2);
				edges.pop();
			}
		}
	}

	protected void setLocationSeq(Stack<Edge> edges, final Block seqNode, int x, final int y) {
		Dim dim = dimensions.get(edges);

		if (!edges.isEmpty() && (edges.peek().isBlockable() || edges.peek().isHideable())) {
			if (dim.isLeft2right()) {
				x += CONFIGSIZE;
			} else {
				x -= CONFIGSIZE;
			}
		}

		if (!dim.isLeft2right()) {
			// find the last child.
			int w = 0;
			for (Edge e : seqNode.getOutgoingEdges()) {
				edges.push(e);
				w += dimensions.get(edges).width;
				edges.pop();
			}
			x += w;

		}

		for (Edge e : seqNode.getOutgoingEdges()) {
			edges.push(e);
			if (!dim.isLeft2right()) {
				x -= dimensions.get(edges).width;
			}
			setLocation(edges, x, y + (dim.height - dimensions.get(edges).height) / 2);
			if (dim.isLeft2right()) {
				x += dimensions.get(edges).width;
			}
			edges.pop();
		}
	}

	protected void setLocationOther(Stack<Edge> edges, final Block otherNode, int x, int y) {
		// order of the nodes does not matter. Can we use that?
		Dim dim = dimensions.get(edges);

		x += OPERATORWIDTH;
		for (Edge e : otherNode.getOutgoingEdges()) {
			edges.push(e);
			setLocation(edges, x + (dim.width - 2 * OPERATORWIDTH - dimensions.get(edges).width) / 2, y);
			y += dimensions.get(edges).height;
			edges.pop();
		}
	}

	public JGraph getJGraph() {
		GraphModel model = new DefaultGraphModel();
		GraphLayoutCache view = new GraphLayoutCache(model, new TreeCellViewFactory());
		final JGraph graph = new JGraph(model, view);
		graph.setAntiAliased(true);
		graph.setScale(1.0);
		drawNode(graph.getGraphLayoutCache(), null, null);
		return graph;
	}

	protected SeSeGroup drawNode(GraphLayoutCache g, DefaultGraphCell parent, Stack<Edge> edges) {

		if (parent == null) {
			// ensure the entire tree is drawn first
			DefaultGraphCell newParent = new DefaultGraphCell(new Path());
			SeSeGroup ports = drawNode(g, newParent, new Stack<Edge>());

			Dim dim = dimensions.get(null);
			SeSeGroup inPort = drawOperator(g, newParent, 0, (dim.height - EVENTHEIGHT) / 2, OperatorType.START);
			SeSeGroup outPort = drawOperator(g, newParent, dim.width - EVENTWIDTH, (dim.height - EVENTHEIGHT) / 2,
					OperatorType.END);
			Expression ex = null;
			if (tree.getRoot() instanceof Block) {
				ex = ((Block) tree.getRoot()).getOutgoingEdges().get(0).getExpression();
			}

			if (ports != null) {
				drawEdge(g, newParent, inPort.getIn(), ports.getIn(), EdgeRouting.VERTICAL_FIRST, 0.0, false, false,
						false, ex);
				drawEdge(g, newParent, ports.getOut(), outPort.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0, false,
						false, false, null);
			} else {
				drawEdge(g, newParent, inPort.getIn(), outPort.getOut(), EdgeRouting.NORMAL, 0.0, false, false, false,
						null);
			}

			g.insert(newParent);
			return new SeSeGroup(inPort.getIn(), outPort.getOut());
		}

		Node node;
		if (edges.isEmpty()) {
			node = tree.getRoot();
		} else {
			node = edges.peek().getTarget();
		}
		if(node == null){
			return new SeSeGroup(new DefaultPort(), new DefaultPort());
		}
		
		DefaultGraphCell newParent = null;
		if (tree.getType(node) != Type.MANTASK && tree.getType(node) != Type.AUTOTASK) {
			Path path = new Path(edges);
			newParent = new DefaultGraphCell(path);
			parent.add(newParent);
			GraphConstants.setBounds(newParent.getAttributes(), dimensions.get(path).toRectangle2D());
		}
		// set the location of this node
		switch (tree.getType(node)) {
			case LOOPXOR :
			case LOOPDEF :
				return drawNodeLoop(g, newParent, edges, (Block) node);
			case SEQ :
				return drawNodeSeq(g, newParent, edges, (Block) node);
			case AND :
			case OR :
			case DEF :
			case XOR :
			case PLACEHOLDER :
				return drawNodeOther(g, newParent, edges, (Block) node);
			case MESSAGE :
			case TIMEOUT :
				return drawNodeEvent(g, newParent, edges, (Event) node);
			case AUTOTASK :
			case MANTASK :
			default :
				return drawNodeLeaf(g, parent, edges, (Task) node);
		}

	}

	protected SeSeGroup drawNodeOther(GraphLayoutCache g, DefaultGraphCell parent, Stack<Edge> edges, Block node) {
		Dim d = new Dim(dimensions.get(edges));
		if (!edges.isEmpty() && (edges.peek().isBlockable() || edges.peek().isHideable())) {
			if (d.isLeft2right()) {
				d.x += CONFIGSIZE;
				d.width -= CONFIGSIZE;
			} else {
				d.x -= CONFIGSIZE;
				d.width -= CONFIGSIZE;
			}
		}

		OperatorType t1, t2;
		switch (tree.getType(node)) {
			case XOR :
				t1 = OperatorType.XOR;
				t2 = t1;
				break;
			case DEF :
				t1 = OperatorType.DEF;
				t2 = OperatorType.XOR;
				break;
			case AND :
				t1 = OperatorType.AND;
				t2 = t1;
				break;
			case OR :
				t1 = OperatorType.OR;
				t2 = t1;
				break;
			case PLACEHOLDER :
				t1 = OperatorType.PLACEHOLDER;
				t2 = t1;
				break;
			case LOOPXOR :
			case LOOPDEF :
			case SEQ :
			default :
				t1 = OperatorType.XOR;
				t2 = t1;
		}

		//		if (tree.getType(node) == Type.AND) {
		//			t = OperatorType.AND;
		//		} else if (tree.getType(node) == ProcessTree.OR) {
		//			t = OperatorType.OR;
		//		} else { // (tree.getTypeFast(node) == ProcessTree.XOR) 
		//			t = OperatorType.XOR;
		//		}

		SeSeGroup portIn, portOut;
		int opY = d.y + (d.height - OPERATORHEIGHT) / 2;
		if (d.isLeft2right()) {
			portIn = drawOperator(g, parent, d.x, opY, t1);
			portOut = drawOperator(g, parent, d.x + d.width - OPERATORWIDTH, opY, t2);
		} else {
			portOut = drawOperator(g, parent, d.x, opY, t1);
			portIn = drawOperator(g, parent, d.x + d.width - OPERATORWIDTH, opY, t2);
		}
		opY += OPERATORHEIGHT / 2;

		TIntList addSkipEdgeY = new TIntArrayList();
		boolean edgeAdded = false;
		SeSeGroup ports;
		for (Edge e : node.getOutgoingEdges()) {
			edges.push(e);
			ports = drawNode(g, parent, edges);
			if (ports == null) {
				Dim dim = dimensions.get(edges);
				addSkipEdgeY.add(dim.y);
				edges.pop();
				continue;
			}
			edgeAdded = true;
			int y = dimensions.get(edges).y + dimensions.get(edges).height / 2;
			if (y != d.y + d.height / 2) {
				drawEdge(g, parent, portIn.getOut(), ports.getIn(), EdgeRouting.VERTICAL_FIRST, 0.0, e.isBlockable(),
						e.isHideable(), t1 == OperatorType.PLACEHOLDER, e.getExpression());
				drawEdge(g, parent, ports.getOut(), portOut.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0, false, false,
						t1 == OperatorType.PLACEHOLDER, null);
			} else {
				drawEdge(g, parent, portIn.getOut(), ports.getIn(), EdgeRouting.VERTICAL_FIRST, 0.0, e.isBlockable(),
						e.isHideable(), t1 == OperatorType.PLACEHOLDER, e.getExpression());
				drawEdge(g, parent, ports.getOut(), portOut.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0, false, false,
						t1 == OperatorType.PLACEHOLDER, null);
			}
			edges.pop();
		}
		if (!edgeAdded) {
			return null;
		}
		if (!addSkipEdgeY.isEmpty()) {
			for (int i = 0; i < addSkipEdgeY.size(); i++) {
				int y = addSkipEdgeY.get(i);
				drawEdge(g, parent, portIn.getIn(), portOut.getOut(), EdgeRouting.SKIP, y - opY, false, false,
						t1 == OperatorType.PLACEHOLDER, null);
			}
		}

		return new SeSeGroup(portIn.getIn(), portOut.getOut());
	}

	protected SeSeGroup drawNodeEvent(GraphLayoutCache g, DefaultGraphCell parent, Stack<Edge> edges, Event node) {
		Dim d = dimensions.get(edges);

		DefaultGraphCell graphCell;
		if (node instanceof Event.TimeOut) {
			graphCell = new DefaultGraphCell(EventType.TIMEOUT);
		} else {
			graphCell = new DefaultGraphCell(EventType.MESSAGE);
		}

		if (d.isLeft2right()) {
			GraphConstants.setBounds(graphCell.getAttributes(), new Rectangle2D.Double(d.x + BORDER, d.y + d.height / 2
					- EVENTHEIGHT / 2 + BORDER, EVENTWIDTH - 2 * BORDER, EVENTHEIGHT - 2 * BORDER));
		} else {
			GraphConstants.setBounds(graphCell.getAttributes(), new Rectangle2D.Double(d.x + d.width - EVENTWIDTH
					- BORDER, d.y + d.height / 2 - EVENTHEIGHT / 2 + BORDER, EVENTWIDTH - 2 * BORDER, EVENTHEIGHT - 2
					* BORDER));
		}
		GraphConstants.setBorder(graphCell.getAttributes(), BorderFactory.createLineBorder(Color.BLACK, 1));
		GraphConstants.setConnectable(graphCell.getAttributes(), false);
		GraphConstants.setDisconnectable(graphCell.getAttributes(), false);
		CellConstants.setVertexShape(graphCell.getAttributes(), MultiLineVertexRenderer.SHAPE_CIRCLE);
		CellConstants.setLineWidth(graphCell.getAttributes(), 1);

		DefaultPort inPort = new DefaultPort();
		graphCell.add(inPort);
		inPort.setParent(graphCell);

		DefaultPort outPort = new DefaultPort();
		graphCell.add(outPort);
		outPort.setParent(graphCell);

		GraphConstants.setOffset(inPort.getAttributes(), new Point2D.Double((d.isLeft2right() ? 0
				: GraphConstants.PERMILLE), GraphConstants.PERMILLE / 2));
		GraphConstants.setOffset(outPort.getAttributes(), new Point2D.Double(
				(d.isLeft2right() ? GraphConstants.PERMILLE : 0), GraphConstants.PERMILLE / 2));

		parent.add(graphCell);
		//g.insert(graphCell);

		Edge out = node.getOutgoingEdges().get(0);
		edges.push(out);
		SeSeGroup ports = drawNode(g, parent, edges);
		edges.pop();

		if (ports != null) {
			drawEdge(g, parent, outPort, ports.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0, false, false, false,
					out.getExpression());

			return new SeSeGroup(inPort, ports.getOut());
		} else {
			return new SeSeGroup(inPort, outPort);
		}
	}

	protected SeSeGroup drawNodeLeaf(GraphLayoutCache g, DefaultGraphCell parent, Stack<Edge> edges, Task node) {
		Dim d = dimensions.get(edges);

		if (node instanceof Task.Automatic) {
			Automatic auto = (Automatic) node;
			if (auto.getReadVariables().isEmpty() && auto.getWrittenVariables().isEmpty()) {
				return null;
			}
		}

		DefaultGraphCell graphCell;
		graphCell = new DefaultGraphCell(node.toStringShort());

		GraphConstants.setBounds(graphCell.getAttributes(), new Rectangle2D.Double(d.x + BORDER, d.y + BORDER, d.width
				- 2 * BORDER, d.height - 2 * +BORDER));
		GraphConstants.setBorder(graphCell.getAttributes(), BorderFactory.createLineBorder(Color.BLACK, 2));
		GraphConstants.setConnectable(graphCell.getAttributes(), false);
		GraphConstants.setDisconnectable(graphCell.getAttributes(), false);
		CellConstants.setVertexShape(graphCell.getAttributes(), MultiLineVertexRenderer.SHAPE_ROUNDED);
		CellConstants.setLineWidth(graphCell.getAttributes(), 1);

		if (node instanceof Task.Automatic) {
			//			GraphConstants.setHorizontalAlignment(graphCell.getAttributes(), SwingConstants.LEFT);
			GraphConstants.setVerticalAlignment(graphCell.getAttributes(), SwingConstants.TOP);
			GraphConstants.setVerticalTextPosition(graphCell.getAttributes(), SwingConstants.BOTTOM);
			//			GraphConstants.setHorizontalTextPosition(graphCell.getAttributes(), SwingConstants.CENTER);
			GraphConstants.setIcon(graphCell.getAttributes(), new Icon() {

				public void paintIcon(Component c, Graphics g, int x, int y) {
					g.drawOval(x, y + 2, 8, 8);
					g.drawOval(x + 10, y + 9, 8, 8);
					g.drawOval(x + 2, y + 4, 4, 4);
					g.drawOval(x + 12, y + 11, 4, 4);
				}

				public int getIconWidth() {
					return 20;
				}

				public int getIconHeight() {
					return 20;
				}

			});
		}

		DefaultPort inPort = new DefaultPort();
		graphCell.add(inPort);
		inPort.setParent(graphCell);

		DefaultPort outPort = new DefaultPort();
		graphCell.add(outPort);
		outPort.setParent(graphCell);

		GraphConstants.setOffset(inPort.getAttributes(), new Point2D.Double((d.isLeft2right() ? 0
				: GraphConstants.PERMILLE), GraphConstants.PERMILLE / 2));
		GraphConstants.setOffset(outPort.getAttributes(), new Point2D.Double(
				(d.isLeft2right() ? GraphConstants.PERMILLE : 0), GraphConstants.PERMILLE / 2));

		parent.add(graphCell);
		
		int offset = 0;
		for(ExtraInfoRenderer ren: renderers){
			Set<DefaultGraphCell> cells = new HashSet<DefaultGraphCell>();
			try {
				cells = ren.drawComponent(g, node.getID(), d.x + 0.5* BORDER, d.y + BORDER + LEAFHEIGHT - 2 * BORDER + 10 + offset, PERFORMANCEWIDTH, (PERFORMANCEHEIGHT - 10) / renderers.size());
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			for(DefaultGraphCell cell: cells){
				parent.add(cell);
			}
			offset += PERFORMANCEHEIGHT / renderers.size();
		}
		
		//		g.insert(graphCell);
		return new SeSeGroup(inPort, outPort);
	}

	protected SeSeGroup drawNodeLoop(GraphLayoutCache g, DefaultGraphCell parent, Stack<Edge> edges, Block loopNode) {
		Dim d = new Dim(dimensions.get(edges));
		if (!edges.isEmpty() && (edges.peek().isBlockable() || edges.peek().isHideable())) {
			if (d.isLeft2right()) {
				d.x += CONFIGSIZE;
				d.width -= CONFIGSIZE;
			} else {
				d.x -= CONFIGSIZE;
				d.width -= CONFIGSIZE;
			}
		}

		Edge left = loopNode.getOutgoingEdges().get(0);
		Edge middle = loopNode.getOutgoingEdges().get(1);
		Edge right = loopNode.getOutgoingEdges().get(2);
		edges.push(right);
		List<Edge> rightIndex = new ArrayList<Edge>(edges);
		edges.pop();

		int xi1, xi2;
		if (d.isLeft2right()) {
			xi1 = d.x;
			xi2 = d.x + d.width - dimensions.get(rightIndex).width - OPERATORWIDTH;
		} else {
			xi1 = d.x + d.width - OPERATORWIDTH;
			xi2 = d.x + dimensions.get(rightIndex).width;
		}

		int cy = d.y + (d.height - OPERATORHEIGHT) / 2;

		SeSeGroup port1 = drawOperator(g, parent, xi1, cy, OperatorType.XOR);
		SeSeGroup port2 = drawOperator(g, parent, xi2, cy, loopNode instanceof Block.XorLoop ? OperatorType.XOR
				: OperatorType.DEF);

		edges.push(left);
		SeSeGroup leftPorts = drawNode(g, parent, edges);
		edges.pop();
		edges.push(middle);
		SeSeGroup middlePorts = drawNode(g, parent, edges);
		edges.pop();
		edges.push(right);
		SeSeGroup rightPorts = drawNode(g, parent, edges);
		edges.pop();

		if (leftPorts != null) {
			drawEdge(g, parent, port1.getOut(), leftPorts.getIn(), EdgeRouting.VERTICAL_FIRST, 0.0, left.isBlockable(),
					left.isHideable(), false, left.getExpression());
			drawEdge(g, parent, leftPorts.getOut(), port2.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0, false, false,
					false, null);
		} else {
			drawEdge(g, parent, port1.getOut(), port2.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0, false, false, false,
					null);
		}

		if (middlePorts != null) {
			drawEdge(g, parent, port2.getOut(), middlePorts.getIn(), EdgeRouting.VERTICAL_FIRST, 0.0,
					middle.isBlockable(), middle.isHideable(), false, middle.getExpression());
			drawEdge(g, parent, middlePorts.getOut(), port1.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0, false, false,
					false, null);
		} else {
			drawEdge(g, parent, port2.getOut(), port1.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0, false, false, false,
					null);
		}

		if (rightPorts != null) {
			drawEdge(g, parent, port2.getOut(), rightPorts.getIn(), EdgeRouting.HORIZONTAL_FIRST, 0.0,
					right.isBlockable(), right.isHideable(), false, right.getExpression());

			return new SeSeGroup(port1.getIn(), rightPorts.getOut());
		} else {
			return new SeSeGroup(port1.getIn(), port2.getOut());
		}
	}

	protected SeSeGroup drawOperator(GraphLayoutCache g, DefaultGraphCell parent, int x, int y, OperatorType type) {

		DefaultGraphCell graphCell = new DefaultGraphCell(type);
		GraphConstants.setSizeable(graphCell.getAttributes(), false);
		GraphConstants.setResize(graphCell.getAttributes(), false);
		GraphConstants.setBounds(graphCell.getAttributes(), new Rectangle2D.Double(x + BORDER, y + BORDER,
				OPERATORWIDTH - 2 * BORDER, OPERATORHEIGHT - 2 * BORDER));
		DefaultPort port = new DefaultPort();
		graphCell.add(port);
		port.setParent(graphCell);
		GraphConstants.setBorder(graphCell.getAttributes(), BorderFactory.createLineBorder(Color.BLACK, 1));

		GraphConstants.setConnectable(graphCell.getAttributes(), false);
		GraphConstants.setDisconnectable(graphCell.getAttributes(), false);
		if (type == OperatorType.END || type == OperatorType.START) {
			GraphConstants.setBounds(graphCell.getAttributes(), new Rectangle2D.Double(x + BORDER, y + BORDER,
					EVENTWIDTH - 2 * BORDER, EVENTWIDTH - 2 * BORDER));
			CellConstants.setVertexShape(graphCell.getAttributes(), MultiLineVertexRenderer.SHAPE_CIRCLE);
		} else if (type == OperatorType.PLACEHOLDER) {
			GraphConstants.setBounds(graphCell.getAttributes(), new Rectangle2D.Double(x + BORDER, y + BORDER,
					OPERATORWIDTH - 2 * BORDER, OPERATORHEIGHT - 2 * BORDER));
			CellConstants.setVertexShape(graphCell.getAttributes(), MultiLineVertexRenderer.SHAPE_CIRCLE);
			CellConstants.setLineWidth(graphCell.getAttributes(), 0);
			CellConstants.setBorder(graphCell.getAttributes(), BorderFactory.createEmptyBorder());
		} else {
			CellConstants.setVertexShape(graphCell.getAttributes(), MultiLineVertexRenderer.SHAPE_DIAMOND);
		}
		if (type == OperatorType.END) {
			CellConstants.setLineWidth(graphCell.getAttributes(), 3);
		} else if (type != OperatorType.PLACEHOLDER) {
			CellConstants.setLineWidth(graphCell.getAttributes(), 1);
		}
		if (parent != null) {
			parent.add(graphCell);
		} else {
			g.insert(graphCell);
		}
		return new SeSeGroup(port, port);
	}

	@SuppressWarnings("unchecked")
	protected void drawEdge(GraphLayoutCache g, DefaultGraphCell parent, DefaultPort from, DefaultPort to,
			EdgeRouting routing, double routeYAt, boolean blockable, boolean hideable, boolean dashed, Expression label) {
		DefaultEdge edge;
		edge = new DefaultEdge(label);
		edge.getAttributes().put(BLOCKABLE, blockable);
		edge.getAttributes().put(HIDEABLE, hideable);

		edge.setSource(from);
		edge.setTarget(to);

		edge.getAttributes().put(EDGEROUTING, routing);
		if (routing == EdgeRouting.SKIP) {
			edge.getAttributes().put(ROUTESKIPY, routeYAt);
		}
		// TODO: massive hack
		//GraphConstants.setLabelEnabled(edge.getAttributes(), (label != null));
		GraphConstants.setLineEnd(edge.getAttributes(), GraphConstants.ARROW_TECHNICAL);
		GraphConstants.setEndSize(edge.getAttributes(), ARROWSIZE);
		GraphConstants.setEndFill(edge.getAttributes(), true);
		GraphConstants.setConnectable(edge.getAttributes(), false);
		GraphConstants.setDisconnectable(edge.getAttributes(), false);
		GraphConstants.setLineWidth(edge.getAttributes(), 1);

		if (dashed) {
			CellConstants.setDashPattern(edge.getAttributes(), new float[] { 7F, 7F });
		}

		GraphConstants.setRouting(edge.getAttributes(), ROUTING);
		//GraphConstants.setRouting(edge.getAttributes(), GraphConstants.ROUTING_DEFAULT);

		GraphConstants.setLineStyle(edge.getAttributes(), STYLE);
		if (parent != null) {
			parent.add(edge);
		} else {
			g.insert(edge);
		}

		//System.out.println(GraphConstants.getPoints(edge.getAttributes()));

	}

	protected SeSeGroup drawNodeSeq(GraphLayoutCache g, DefaultGraphCell parent, Stack<Edge> edges, Block node) {
		SeSeGroup ports;
		DefaultPort in = null;
		DefaultPort out = null;
		for (Edge e : node.getOutgoingEdges()) {
			edges.push(e);
			ports = drawNode(g, parent, edges);
			edges.pop();
			if (ports != null) {
				if (in == null) {
					in = ports.getIn();
				} else {
					drawEdge(g, parent, out, ports.getIn(), EdgeRouting.NORMAL, 0.0, e.isBlockable(), e.isHideable(),
							false, e.getExpression());
				}
				out = ports.getOut();
			}
		}
		if (in == null) {
			return null;
		}
		return new SeSeGroup(in, out);
	}

	public static void main(String[] args) {

		ProcessTree tree = TestProcessTree.getTree();

		TreeBPMNLayoutBuilder builder = new TreeBPMNLayoutBuilder(tree, null);
		JFrame frame = new JFrame("Test of TreeBuilder");
		JGraph graph = builder.getJGraph();
		JScrollPane scroll = new JScrollPane(graph);
		frame.add(scroll);
		frame.setSize(builder.getDimension(null));
		frame.setPreferredSize(builder.getDimension(null));
		frame.pack();
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setVisible(true);

	}

	public Dimension getDimension(List<Edge> path) {
		return new Dimension(dimensions.get(path).width + 40, dimensions.get(path).height + 50);
	}

}
