package org.processmining.processtree.visualization.tree;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.SwingConstants;

import org.deckfour.xes.classification.XEventClasses;
import org.jgraph.JGraph;
import org.jgraph.graph.DefaultEdge;
import org.jgraph.graph.DefaultGraphCell;
import org.jgraph.graph.DefaultGraphModel;
import org.jgraph.graph.DefaultPort;
import org.jgraph.graph.GraphConstants;
import org.jgraph.graph.GraphLayoutCache;
import org.jgraph.graph.GraphModel;
import org.processmining.processtree.Block;
import org.processmining.processtree.Block.PlaceHolder;
import org.processmining.processtree.Edge;
import org.processmining.processtree.Node;
import org.processmining.processtree.ProcessTree;
import org.processmining.processtree.Task;
import org.processmining.processtree.test.TestProcessTree;
import org.processmining.processtree.visualization.Dim;
import org.processmining.processtree.visualization.extrainforenderer.ExtraInfoRenderer;
import org.processmining.processtree.visualization.jgraph.TreeCellViewFactory;

import com.jgraph.components.labels.CellConstants;
import com.jgraph.components.labels.MultiLineVertexRenderer;

public class TreeLayoutBuilder {

	protected ProcessTree tree;
	protected XEventClasses classes;
	protected Map<Node, Dim> dimensions;
	protected Map<Node, DefaultPort> ports;
	protected List<ExtraInfoRenderer> renderers;
	protected Set<UUID> foldedElements;
	
	public static final int PERFORMANCEWIDTH = 100;
	public static final int PERFORMANCEHEIGHT = 100;

	public static final int BORDER = 24;
	public static final int LEAFHEIGHT = 50 + 2 * BORDER;
	public int LEAFWIDTH = 80 + 2 * BORDER;
	public int OPERATORHEIGHT = 30;
	public int OPERATORWIDTH = 30;
	public static final int ARROWSIZE = 7;
	public static final int CONFIGSIZE = 3 * ARROWSIZE;
	private static final int POINTERWIDTH = 10;
	private static final Dim EMPTY = new Dim(0, POINTERWIDTH, true);
	public static final String BLOCKABLE = "Blockable";
	public static final String HIDEABLE = "Hideable";
	public static final String ID = "id";

	public TreeLayoutBuilder(ProcessTree tree) {
		this(tree, null);
	}

	public TreeLayoutBuilder(ProcessTree tree, XEventClasses classes) {
		this.tree = tree;
		this.classes = classes;
		this.dimensions = new HashMap<Node, Dim>(tree.size());
		this.ports = new HashMap<Node, DefaultPort>(tree.size());
		this.renderers = new ArrayList<ExtraInfoRenderer>();
		this.foldedElements = new HashSet<UUID>();
		buildLayout();
	}
	
	public TreeLayoutBuilder(ProcessTree tree, List<ExtraInfoRenderer> renderers, XEventClasses classes) {
		this.tree = tree;
		this.renderers = renderers;
		this.classes = classes;
		this.dimensions = new HashMap<Node, Dim>(tree.size());
		this.ports = new HashMap<Node, DefaultPort>(tree.size());
		if(this.renderers == null){
			this.renderers = new ArrayList<ExtraInfoRenderer>();
		}
		if(this.renderers.size() > 0){
			OPERATORHEIGHT = PERFORMANCEHEIGHT;
			OPERATORWIDTH = PERFORMANCEWIDTH;
			LEAFWIDTH = Math.max(80 + 2 * BORDER, PERFORMANCEWIDTH + 2 * BORDER);
		}
		this.foldedElements = new HashSet<UUID>();
		buildLayout();
	}
	
	public TreeLayoutBuilder(ProcessTree tree, List<ExtraInfoRenderer> renderers, XEventClasses classes, Set<UUID> foldedElements) {
		this.tree = tree;
		this.renderers = renderers;
		this.classes = classes;
		this.dimensions = new HashMap<Node, Dim>(tree.size());
		this.ports = new HashMap<Node, DefaultPort>(tree.size());
		if(this.renderers == null){
			this.renderers = new ArrayList<ExtraInfoRenderer>();
		}
		if(this.renderers.size() > 0){
			OPERATORHEIGHT = PERFORMANCEHEIGHT;
			OPERATORWIDTH = PERFORMANCEWIDTH;
			LEAFWIDTH = Math.max(80 + 2 * BORDER, PERFORMANCEWIDTH + 2 * BORDER);
		}
		this.foldedElements = foldedElements;
		buildLayout();
	}

	public JGraph getJGraph() {
		GraphModel model = new DefaultGraphModel();
		GraphLayoutCache view = new GraphLayoutCache(model, new TreeCellViewFactory());
		final JGraph graph = new JGraph(model, view);
		graph.setAntiAliased(true);
		graph.setScale(1.0);
		drawNode(graph.getGraphLayoutCache(), null, tree.getRoot());
		return graph;
	}
	
	public void reset(){
		this.ports = new HashMap<Node, DefaultPort>(tree.size());
	}

	public static void main(String[] args) {

		ProcessTree tree = TestProcessTree.getTree();

		TreeLayoutBuilder builder = new TreeLayoutBuilder(tree, null);
		JFrame frame = new JFrame("Test of TreeBuilder");
		JGraph graph = builder.getJGraph();
		JScrollPane scroll = new JScrollPane(graph);
		frame.add(scroll);
		frame.setSize(builder.getDimension(tree.getRoot()));
		frame.setPreferredSize(builder.getDimension(tree.getRoot()));
		frame.pack();
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setVisible(true);

	}

	private Dimension getDimension(Node node) {
		return new Dimension(dimensions.get(node).width + 40, dimensions.get(node).height + 50);
	}

	protected void buildLayout() {
		getSize(tree.getRoot());
		setLocation(tree.getRoot(), 0, 0);
	}

	protected Dim getSize(Node node) {
		if (dimensions.containsKey(node)) {
			return EMPTY;
		}
		// special case for the editor
		if(node == null){
			return EMPTY;
		}
		// a node consists of an operator and potentially children of which the size matters only
		// if they do not have a size yet.
		if (node.isLeaf()) {
			Dim dim = new Dim(LEAFWIDTH, LEAFHEIGHT, false);
			dimensions.put(node, dim);
			return dim;
		}

		int height = 0;
		int width = 0;

		boolean cfg = false;
		if(!foldedElements.contains(node.getID())){
			for (Edge edge : ((Block) node).getOutgoingEdges()) {
				cfg |= edge.isBlockable() || edge.isHideable();
				Dim s = getSize(edge.getTarget());
				// set the width to the width of the operator if it is in the folded nodes
				width += s.width;
				height = Math.max(height, s.height);
			}
		}

		Dim dim = new Dim(Math.max(OPERATORWIDTH, width), height + OPERATORHEIGHT + BORDER + (cfg ? CONFIGSIZE : 0),
				false);
		dimensions.put(node, dim);
		return dim;
	}

	protected void setLocation(Node node, int x, int y) {
		// special case for the editor
		if(node  == null){
			return;
		}
		
		Dim dim = dimensions.get(node);
		dim.setLocation(x, y);
		dim.setLeft2right(true);

		if (node instanceof Block) {
			// reserve space for operator
			y += OPERATORHEIGHT + BORDER;
			// allow space for configuration icon
			for (Edge edge : ((Block) node).getOutgoingEdges()) {
				if (edge.isBlockable() || edge.isHideable()) {
					y += CONFIGSIZE;
					break;
				}
			}
			if(!foldedElements.contains(node.getID())){
				for (Edge edge : ((Block) node).getOutgoingEdges()) {
					Node ch = edge.getTarget();
					Dim chDim = dimensions.get(ch);
					if (!chDim.isLeft2right()) {
						chDim.setLeft2right(true);
						setLocation(ch, x, y);
						x += chDim.width;
					} else {
						x += POINTERWIDTH;
					}
				}
			}
		}
	}

	protected DefaultPort drawNode(GraphLayoutCache g, DefaultGraphCell parent, Node node) {
		// special case for the editor
		if(node == null){
			return new DefaultPort();
		}
		boolean draw = false;
		for(Edge e: node.getIncomingEdges()){
			if(!foldedElements.contains(e.getID())){
				draw = true;
			}
		}
		draw = draw || tree.getRoot().equals(node);
		if (draw && ports.containsKey(node)) {
			return ports.get(node);
		}
		DefaultGraphCell newParent = null;
		if(draw){
			DefaultPort port;
			if (node instanceof Block) {
				//newParent = new DefaultGraphCell(node);
				//GraphConstants.setBounds(newParent.getAttributes(), dimensions.get(node).toRectangle2D());
				port = drawBlock(g, newParent, (Block) node);
				//g.insert(newParent);
			} else {
				port = drawLeaf(g, parent, node);
			}
			return port;
		}
		else{
			return null;
		}
	}

	@SuppressWarnings("unchecked")
	protected DefaultPort drawLeaf(GraphLayoutCache g, DefaultGraphCell parent, Node node) {
		boolean draw = false;
		for(Edge e: node.getIncomingEdges()){
			if(!foldedElements.contains(e.getSource().getID())){
				draw = true;
			}
		}
		// special case with only a leaf as the root
		if(node.getIncomingEdges().isEmpty()){
			draw = true;
		}
		if(!draw){
			foldedElements.add(node.getID());
			return null;
		}
		else{
			foldedElements.remove(node.getID());
		}
		Dim d = dimensions.get(node);

		DefaultGraphCell graphCell;
		graphCell = new DefaultGraphCell(node.toStringShort());

		GraphConstants.setBounds(graphCell.getAttributes(), new Rectangle2D.Double(d.x + BORDER, d.y + BORDER, d.width
				- 2 * BORDER, d.height - 2 * +BORDER));
		GraphConstants.setBorder(graphCell.getAttributes(), BorderFactory.createLineBorder(Color.BLACK, 2));
		GraphConstants.setConnectable(graphCell.getAttributes(), false);
		GraphConstants.setDisconnectable(graphCell.getAttributes(), false);
		GraphConstants.setVerticalAlignment(graphCell.getAttributes(), SwingConstants.TOP);
		CellConstants.setVertexShape(graphCell.getAttributes(), MultiLineVertexRenderer.SHAPE_ROUNDED);
		CellConstants.setLineWidth(graphCell.getAttributes(), 1);
		graphCell.getAttributes().put(ID, node.getID());

		if (node instanceof Task.Automatic) {
			GraphConstants.setBackground(graphCell.getAttributes(), Color.BLACK);
			GraphConstants.setOpaque(graphCell.getAttributes(), true);

		}

		DefaultPort inPort = new DefaultPort();
		graphCell.add(inPort);
		inPort.setParent(graphCell);

		GraphConstants.setOffset(inPort.getAttributes(), new Point2D.Double(GraphConstants.PERMILLE / 2, 0));
		g.insert(graphCell);
		ports.put(node, inPort);
		double offsetVis = 0;
		for(ExtraInfoRenderer ren: renderers){
			Set<DefaultGraphCell> cells = new HashSet<DefaultGraphCell>();
			try {
				cells = ren.drawComponent(g, node.getID(), d.x + 0.5* BORDER, d.y + BORDER + LEAFHEIGHT - 2 * BORDER + 10 + offsetVis, PERFORMANCEWIDTH, (PERFORMANCEHEIGHT - 10) / renderers.size());
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			for(DefaultGraphCell cell: cells){
				if(draw){
					g.insert(cell);
				}
			}
			offsetVis += PERFORMANCEHEIGHT / renderers.size();
		}

		return inPort;
	}

	@SuppressWarnings("unchecked")
	protected DefaultPort drawBlock(GraphLayoutCache g, DefaultGraphCell parent, Block node) {
		boolean draw = false;
		for(Edge e: node.getIncomingEdges()){
			if(!foldedElements.contains(e.getSource().getID())){
				draw = true;
			}
		}
		draw = draw || tree.getRoot().equals(node);
		if(!draw){
			foldedElements.add(node.getID());
			return null;
		}
		Dim d = dimensions.get(node);

		DefaultGraphCell graphCell = new DefaultGraphCell(ProcessTree.Type.getTypeOf(node));
		GraphConstants.setBounds(graphCell.getAttributes(), new Rectangle2D.Double(d.x + d.width / 2 - OPERATORWIDTH
				/ 2, d.y + BORDER, OPERATORWIDTH, OPERATORHEIGHT));

		GraphConstants.setBorder(graphCell.getAttributes(), BorderFactory.createLineBorder(Color.BLACK, 1));
		GraphConstants.setBackground(graphCell.getAttributes(), Color.WHITE);
		GraphConstants.setConnectable(graphCell.getAttributes(), false);
		GraphConstants.setDisconnectable(graphCell.getAttributes(), false);
		CellConstants.setVertexShape(graphCell.getAttributes(), MultiLineVertexRenderer.SHAPE_CIRCLE);
		CellConstants.setLineWidth(graphCell.getAttributes(), 1);
		CellConstants.setResize(graphCell.getAttributes(), false);
		CellConstants.setSizeable(graphCell.getAttributes(), false);
		graphCell.getAttributes().put(ID, node.getID());
		
		g.insert(graphCell);

		DefaultPort inPort = new DefaultPort();
		graphCell.add(inPort);
		inPort.setParent(graphCell);

		GraphConstants.setOffset(inPort.getAttributes(), new Point2D.Double(GraphConstants.PERMILLE / 2, 0));

		ports.put(node, inPort);

		double skip = 1.0 / node.getOutgoingEdges().size();
		double offset = skip / 2.0;
		for (Edge e : node.getOutgoingEdges()) {
			if(foldedElements.contains(node.getID())){
				break;
			}
			DefaultPort to = drawNode(g, parent, e.getTarget());

			DefaultPort outPort = new DefaultPort();
			graphCell.add(outPort);
			outPort.setParent(graphCell);

			double y = Math.sqrt(1.0 - (0.5 - offset) * (0.5 - offset));

			GraphConstants.setOffset(outPort.getAttributes(), new Point2D.Double(offset * GraphConstants.PERMILLE, y
					* GraphConstants.PERMILLE));
			drawEdge(g, parent, outPort, to, e, node instanceof PlaceHolder);
			offset += skip;
		}
		
		double offsetVis = 0;
		for(ExtraInfoRenderer ren: renderers){
			HashSet<DefaultGraphCell> cells = new HashSet<DefaultGraphCell>();
			try {
				cells = ren.drawComponent(g, node.getID(), d.x + d.width / 2 - OPERATORWIDTH / 2 + OPERATORWIDTH + 10, d.y + BORDER + offsetVis, PERFORMANCEWIDTH - 10, (PERFORMANCEHEIGHT - 10) / renderers.size());
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			for(DefaultGraphCell cell: cells){
				//parent.add(cell);
				g.insert(cell);
			}
			offsetVis += PERFORMANCEHEIGHT / renderers.size();
		}

		return inPort;
	}

	@SuppressWarnings("unchecked")
	protected void drawEdge(GraphLayoutCache g, DefaultGraphCell parent, DefaultPort from, DefaultPort to, Edge e,
			boolean dashed) {
		DefaultEdge edge;
		if(foldedElements.contains(e.getSource())){
			foldedElements.add(e.getID());
		}
		else{
			foldedElements.remove(e.getID());
		}
		if(foldedElements.contains(e.getID())){
			return;
		}
		
		edge = new DefaultEdge(e.getExpression());
		edge.getAttributes().put(BLOCKABLE, e.isBlockable());
		edge.getAttributes().put(HIDEABLE, e.isHideable());
		edge.getAttributes().put(ID, e.getID());

		edge.setSource(from);
		edge.setTarget(to);
		// TODO: massive hack
		//GraphConstants.setLabelEnabled(edge.getAttributes(), (e.getExpression() != null));
		GraphConstants.setLineEnd(edge.getAttributes(), GraphConstants.ARROW_TECHNICAL);
		GraphConstants.setEndSize(edge.getAttributes(), ARROWSIZE);
		GraphConstants.setEndFill(edge.getAttributes(), true);
		GraphConstants.setConnectable(edge.getAttributes(), false);
		GraphConstants.setDisconnectable(edge.getAttributes(), false);
		GraphConstants.setLineWidth(edge.getAttributes(), 1);

		if (dashed) {
			CellConstants.setDashPattern(edge.getAttributes(), new float[] { 7F, 7F });
		}

		//GraphConstants.setRouting(edge.getAttributes(), GraphConstants.ROUTING_SIMPLE);
		//GraphConstants.setLineStyle(edge.getAttributes(), GraphConstants.STYLE_ORTHOGONAL);
		g.insert(edge);
	}

}
