package org.processmining.projectedrecallandprecision.framework;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.plugins.InductiveMiner.Pair;
import org.processmining.projectedrecallandprecision.helperclasses.AutomatonFailedException;
import org.processmining.projectedrecallandprecision.helperclasses.BinomialIteratorThreadSafe;
import org.processmining.projectedrecallandprecision.helperclasses.EfficientLog;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAOperations;
import org.processmining.projectedrecallandprecision.helperclasses.automata.RunAutomaton2;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Minimise;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult.ProjectedMeasuresFailedException;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResultImplBasic;

import gnu.trove.set.hash.THashSet;

public abstract class LogModelFramework<M> {
	/**
	 * 
	 * @param model
	 * @return a set of the activities in the model.
	 */
	public abstract Set<String> getActivitiesModel(M model);

	/**
	 * Transform the model into an automaton. Must be thread-safe. Each activity
	 * name should be replaced with the number in which it appears. I.e., for
	 * process trees: seq(a, b, c) projected on [b, c] becomes an automaton of
	 * seq(0, 1).
	 * 
	 * @param canceller
	 * 
	 * @param data.system
	 * @param names
	 * @return
	 * @throws Exception
	 */
	public abstract DFA projectAndAutomatiseModel(M model, CompareParameters parameters, ProMCanceller canceller,
			String... names) throws Exception;

	/**
	 * 
	 * @param data.system
	 * @param data.k
	 * @return (recall, precision)
	 * @throws AutomatonFailedException
	 * @throws InterruptedException
	 * @throws ProjectedMeasuresFailedException
	 */
	public ProjectedRecallPrecisionResult apply(final M model, final EfficientLog log,
			final CompareParameters parameters, final ProMCanceller canceller)
			throws AutomatonFailedException, InterruptedException, ProjectedMeasuresFailedException {
		boolean multiThreaded = true;

		//make a list of activities
		final Set<String> activitiesLog = getActivitiesLog(log);
		final Set<String> activitiesModel = getActivitiesModel(model);

		Set<String> unodesSet = new THashSet<>(activitiesLog);
		unodesSet.addAll(activitiesModel);
		final List<String> unodes = new ArrayList<>(unodesSet);

		//result keeping
		final ProjectedRecallPrecisionResult result = new ProjectedRecallPrecisionResultImplBasic(parameters.getK(),
				unodes, parameters.getRecallName(), model, null, parameters.isComputeRecallFitness(),
				parameters.isComputePrecision());

		//initialise the iterator
		int threads = parameters.isMultiThreading() ? Math.max(1, Runtime.getRuntime().availableProcessors()) : 1;
		final BinomialIteratorThreadSafe<String> it = new BinomialIteratorThreadSafe<String>(String.class, unodes,
				parameters.getK(), threads);

		//thread pools
		final ExecutorService computationPool = Executors.newFixedThreadPool(threads);
		//final ExecutorService computationPool = Executors.newSingleThreadExecutor();

		//set up threads
		for (int thread = 0; thread < threads; thread++) {
			final int threadNr = thread;
			Runnable job = new Runnable() {
				public void run() {
					String[] combination = it.next(threadNr, parameters.isDebug(), parameters.getDebugEvery());
					while (combination != null && result.getNumberOfFailedMeasures() <= 0) {
						try {
							if (canceller.isCancelled()) {
								result.set(-2, -2, combination);
								return;
							}

							Pair<Double, Double> pair = recallPrecisionOverActivity(model, log, activitiesLog,
									activitiesModel, parameters, canceller, combination);

							if (canceller.isCancelled() || pair == null) {
								result.set(-2, -2, combination);
								return;
							}

							result.set(pair.getA(), pair.getB(), combination);
						} catch (Exception e) {
							e.printStackTrace();
							result.set(-2, -2, combination);
						}
						combination = it.next(threadNr, parameters.isDebug(), parameters.getDebugEvery());
					}
				}
			};
			if (multiThreaded) {
				computationPool.submit(job);
			} else {
				job.run();
			}
		}

		//wait for completion of all threads
		computationPool.shutdown();
		computationPool.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);

		if (canceller.isCancelled()) {
			return null;
		}

		if (result.getNumberOfFailedMeasures() > 0) {
			throw new ProjectedRecallPrecisionResult.ProjectedMeasuresFailedException();
		}

		return result;
	}

	/**
	 * 
	 * @param log
	 * @return a set of activities that appear in the log
	 */
	public static Set<String> getActivitiesLog(EfficientLog log) {
		return new HashSet<String>(Arrays.asList(log.getActivities()));
	}

	public Pair<Double, Double> recallPrecisionOverActivity(M model, EfficientLog log, Set<String> logActivities,
			Set<String> modelActivities, CompareParameters parameters, ProMCanceller canceller, String... names)
			throws Exception {
		DFA automatonModel = projectAndAutomatiseModel(model, parameters, canceller, names);

		if (canceller.isCancelled() || automatonModel == null) {
			return null;
		}

		RunAutomaton2 runAutomatonModel = new RunAutomaton2(automatonModel, canceller);

		if (canceller.isCancelled() || runAutomatonModel == null) {
			return null;
		}

		//check whether the log expresses something about any activity
		/*
		 * { boolean logContainsAny = false; for (int i = 0; i < names.length;
		 * i++) { logContainsAny = logContainsAny ||
		 * logActivities.contains(names[i]); } if (!logContainsAny) { //the
		 * model expresses more than the log; no fitness involved, bad
		 * precision. return Pair.of(-1d, 0d); } }
		 */

		//check whether the model expresses something about any activity
		/*
		 * { boolean modelContainsAny = false; for (int i = 0; i < names.length;
		 * i++) { modelContainsAny = modelContainsAny ||
		 * modelActivities.contains(names[i]); } if (!modelContainsAny) { //the
		 * log expresses more than the model; bad fitness, no precision
		 * involved. return Pair.of(0d, -1d); } }
		 */

		Pair<Integer, DFA> p = processLog(log, runAutomatonModel, canceller, names);
		if (canceller.isCancelled() || p == null) {
			return null;
		}

		DFA automatonLog = p.getB();
		int replayedTraces = p.getA();

		//recall is the portion of traces that is replayable
		double recall = replayedTraces / (log.size() * 1.0);

		double precision = 10000;
		if (parameters.isComputePrecision()) {
			precision = ModelModelFramework.compareAutomata(automatonModel, automatonLog, canceller);

			if (canceller.isCancelled()) {
				return null;
			}

			if (precision > 1) {
				throw new ProjectedMeasuresFailedException();
			}
		}

		if (recall > 1) {
			throw new ProjectedMeasuresFailedException();
		}

		return Pair.of(recall, precision);
	}

	public static Pair<Integer, DFA> processLog(EfficientLog log, RunAutomaton2 modelAutomaton,
			ProMCanceller canceller, String... names) {
		//make a map of names
		short[] projectionKey = log.getProjectionKey(names);

		//compute fitness and precision
		int replayableTraces = 0;
		DFA logAutomaton = new DFAImplEdgeList();
		for (int traceIndex = 0; traceIndex < log.size(); traceIndex++) {

			boolean canReplay = addTraceToAutomaton(modelAutomaton, logAutomaton, log, traceIndex, projectionKey,
					canceller);

			if (canceller.isCancelled()) {
				return null;
			}

			if (canReplay) {
				replayableTraces++;
			}

		}

		Automaton2Minimise.minimise(logAutomaton, canceller);
		if (canceller.isCancelled()) {
			return null;
		}

		return Pair.of(replayableTraces, logAutomaton);
	}

	/**
	 * Add trace to log automaton and returns whether the trace can be replayed
	 * on the model.
	 * 
	 * @param automaton
	 * @param trace
	 * @param canceller
	 * @param classifier
	 * @param namesMap
	 * @return can trace be replayed
	 */
	public static boolean addTraceToAutomaton(RunAutomaton2 modelAutomaton, DFA logAutomaton, EfficientLog log,
			int trace, short[] projectionKey, ProMCanceller canceller) {
		short[] projectedTrace = log.getProjectedTrace(trace, projectionKey);

		//check fitness
		boolean canReplay = modelAutomaton.accepts(projectedTrace);

		//make automation for precision
		DFAOperations.addTrace(logAutomaton, projectedTrace, canceller);

		return canReplay;
	}

	public static String projectedTraceToString(short[] projectedTrace) {
		//make a string
		String result = "";
		for (int event = 0; event < projectedTrace.length; event++) {
			if (projectedTrace[event] >= 0) {
				result += (char) projectedTrace[event];
			}
		}
		return result;
	}
}
