package org.processmining.projectedrecallandprecision.framework;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.plugins.InductiveMiner.Pair;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTreeReduce.ReductionFailedException;
import org.processmining.plugins.InductiveMiner.efficienttree.UnknownTreeNodeException;
import org.processmining.projectedrecallandprecision.helperclasses.AutomatonFailedException;
import org.processmining.projectedrecallandprecision.helperclasses.BinomialIteratorThreadSafe;
import org.processmining.projectedrecallandprecision.helperclasses.MapConjunction2;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList.EdgeIterableOutgoing;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAOperations;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Conjunction;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Minimise;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.StatePair;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResult.ProjectedMeasuresFailedException;
import org.processmining.projectedrecallandprecision.result.ProjectedRecallPrecisionResultImplBasic;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.map.TIntObjectMap;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.THashSet;

public abstract class ModelModelFramework<S, M> {
	/**
	 * 
	 * @param system
	 * @return a set of the activities of the system.
	 */
	public abstract Set<String> getActivitiesSystem(S system);

	/**
	 * 
	 * @param model
	 * @return a set of the activities in the model.
	 */
	public abstract Set<String> getActivitiesModel(M model);

	/**
	 * Transform the system into an automaton. Must be thread-safe. Each
	 * activity name should be replaced with the number in which it appears.
	 * I.e., for process trees: seq(a, b, c) projected on [b, c] becomes an
	 * automaton of seq(0, 1).
	 * 
	 * @param system
	 * @param names
	 * @param canceller
	 * @return
	 * @throws AutomatonFailedException
	 * @throws ReductionFailedException
	 * @throws InterruptedException
	 * @throws UnknownTreeNodeException
	 * @throws CloneNotSupportedException
	 */
	public abstract DFA projectAndAutomatiseSystem(S system, CompareParameters parameters, ProMCanceller canceller,
			String... names) throws AutomatonFailedException, InterruptedException, ReductionFailedException,
			UnknownTreeNodeException, CloneNotSupportedException;

	/**
	 * Transform the model into an automaton. Must be thread-safe. Each activity
	 * name should be replaced with the number in which it appears. I.e., for
	 * process trees: seq(a, b, c) projected on [b, c] becomes an automaton of
	 * seq(0, 1).
	 * 
	 * @param model
	 * @param names
	 * @return
	 * @throws ReductionFailedException
	 * @throws AutomatonFailedException
	 * @throws UnknownTreeNodeException
	 * @throws CloneNotSupportedException
	 */
	public abstract DFA projectAndAutomatiseModel(M model, CompareParameters parameters, ProMCanceller canceller,
			String... names) throws ReductionFailedException, AutomatonFailedException, UnknownTreeNodeException,
			CloneNotSupportedException;

	/**
	 * 
	 * @param system
	 * @param model
	 * @param canceller
	 * @param data.k
	 * @return (recall, precision)
	 * @throws InterruptedException
	 * @throws AutomatonFailedException
	 * @throws ProjectedMeasuresFailedException
	 */
	public ProjectedRecallPrecisionResult apply(S system, M model, CompareParameters parameters,
			ProMCanceller canceller)
			throws AutomatonFailedException, InterruptedException, ProjectedMeasuresFailedException {
		return recallPrecisionOverAllActivities(system, model, parameters, canceller);
	}

	/**
	 * Perform the measurement over all activities.
	 * 
	 * @param system
	 * @param model
	 * @param data.k
	 * @return
	 * @throws InterruptedException
	 * @throws ProjectedMeasuresFailedException
	 * @throws Exception
	 */
	public ProjectedRecallPrecisionResult recallPrecisionOverAllActivities(final S system, final M model,
			final CompareParameters parameters, final ProMCanceller canceller)
			throws AutomatonFailedException, InterruptedException, ProjectedMeasuresFailedException {
		boolean multiThreaded = true;

		//make a list of activities
		final Set<String> activitiesSystem = getActivitiesSystem(system);
		final Set<String> activitiesModel = getActivitiesModel(model);

		Set<String> unodesSet = new THashSet<>(activitiesSystem);
		unodesSet.addAll(activitiesModel);
		final List<String> unodes = new ArrayList<>(unodesSet);

		//result keeping
		final ProjectedRecallPrecisionResult result = new ProjectedRecallPrecisionResultImplBasic(parameters.getK(),
				unodes, parameters.getRecallName(), system, model, parameters.isComputeRecallFitness(),
				parameters.isComputePrecision());

		//initialise the iterator
		int threads = parameters.isMultiThreading() ? Math.max(1, Runtime.getRuntime().availableProcessors()) : 1;
		final BinomialIteratorThreadSafe<String> it = new BinomialIteratorThreadSafe<String>(String.class, unodes,
				parameters.getK(), threads);

		//thread pools
		final ExecutorService computationPool = Executors.newFixedThreadPool(threads);

		//set up threads
		for (int thread = 0; thread < threads; thread++) {
			final int threadNr = thread;
			Runnable r = new Runnable() {
				public void run() {
					String[] combination = it.next(threadNr, parameters.isDebug(), parameters.getDebugEvery());
					while (combination != null && result.getNumberOfFailedMeasures() <= 0) {
						//perform the measurement
						Pair<Double, Double> pair = null;
						try {

							if (canceller.isCancelled()) {
								result.set(-2, -2, combination);
								return;
							}

							pair = recallPrecisionOverActivity(system, model, activitiesSystem, activitiesModel,
									parameters, canceller, combination);

							if (canceller.isCancelled() || pair == null) {
								result.set(-2, -2, combination);
								return;
							}

							result.set(pair.getA(), pair.getB(), combination);
						} catch (Exception e) {
							result.set(-2, -2, combination);
							e.printStackTrace();
						}
						combination = it.next(threadNr, parameters.isDebug(), parameters.getDebugEvery());
					}
				}
			};
			if (multiThreaded) {
				computationPool.submit(r);
			} else {
				r.run();
			}
		}

		//wait for completion of all threads
		computationPool.shutdown();
		computationPool.awaitTermination(Long.MAX_VALUE, TimeUnit.NANOSECONDS);

		if (canceller.isCancelled()) {
			return null;
		}

		if (result.getNumberOfFailedMeasures() > 0) {
			throw new ProjectedRecallPrecisionResult.ProjectedMeasuresFailedException();
		}

		return result;
	}

	/**
	 * Perform the measurement over one n-tuple of activities.
	 * 
	 * @param system
	 * @param model
	 * @param automataCache
	 * @param systemActivities
	 * @param modelActivities
	 * @param names
	 * @param parameters
	 * @return
	 * @throws ReductionFailedException
	 * @throws AutomatonFailedException
	 * @throws InterruptedException
	 * @throws UnknownTreeNodeException
	 * @throws ProjectedMeasuresFailedException
	 * @throws CloneNotSupportedException
	 */
	public Pair<Double, Double> recallPrecisionOverActivity(S system, M model, Set<String> systemActivities,
			Set<String> modelActivities, CompareParameters parameters, ProMCanceller canceller, String... names)
			throws AutomatonFailedException, ReductionFailedException, InterruptedException, UnknownTreeNodeException,
			ProjectedMeasuresFailedException, CloneNotSupportedException {

		//check whether the system expresses something about any activity
		//		{
		//			boolean systemContainsAny = false;
		//			for (int i = 0; i < names.length; i++) {
		//				systemContainsAny = systemContainsAny || systemActivities.contains(names[i]);
		//			}
		//			if (!systemContainsAny) {
		//				//the model expresses more than the system; no fitness involved, bad precision.
		//				return Pair.of(-1d, 0d);
		//			}
		//		}

		//check whether the model expresses something about any activity
		//		{
		//			boolean modelContainsAny = false;
		//			for (int i = 0; i < names.length; i++) {
		//				modelContainsAny = modelContainsAny || modelActivities.contains(names[i]);
		//			}
		//			if (!modelContainsAny) {
		//				//the system expresses more than the model; bad fitness, no precision involved.
		//				return Pair.of(0d, -1d);
		//			}
		//		}

		//compute recall and precision
		DFA automatonSystem = projectAndAutomatiseSystem(system, parameters, canceller, names);

		if (canceller.isCancelled() || automatonSystem == null) {
			return null;
		}

		DFA automatonModel = projectAndAutomatiseModel(model, parameters, canceller, names);

		if (canceller.isCancelled() || automatonModel == null) {
			return null;
		}

		//check whether the automata are equivalent
		//		Automaton.setMinimization(Automaton.MINIMIZE_HOPCROFT);
		//		if (automatonSystem.equals(automatonModel)) {
		//			return Pair.of(1d, 1d);
		//		}
		//disabled as it gives exceptions

		double recall = 100000;
		if (parameters.isComputeRecallFitness()) {
			recall = compareAutomata(automatonSystem, automatonModel, canceller);

			if (canceller.isCancelled()) {
				return null;
			}

			if (recall > 1) {
				throw new ProjectedMeasuresFailedException();
			}
		}

		if (canceller.isCancelled()) {
			return null;
		}

		double precision = 100000;
		if (parameters.isComputePrecision()) {
			precision = compareAutomata(automatonModel, automatonSystem, canceller);

			if (canceller.isCancelled()) {
				return null;
			}

			if (precision > 1) {
				throw new ProjectedMeasuresFailedException();
			}
		}

		if (canceller.isCancelled()) {
			return null;
		}

		return Pair.of(recall, precision);
	}

	/**
	 * 
	 * @param combination
	 * @return whether the strings are unique and sorted
	 */
	public static boolean isWantedCombination(String[] combination) {
		for (int i = 0; i < combination.length - 1; i++) {
			if (combination[i].compareTo(combination[i + 1]) >= 0) {
				return false;
			}
		}
		return true;
	}

	/**
	 * 
	 * @param automatonA
	 * @param automatonB
	 * @param canceller
	 * @return the part of the behaviour in automaton1 that is also in
	 *         automaton2
	 */
	public static double compareAutomata(DFA automatonA, DFA automatonB, ProMCanceller canceller) {
		Pair<DFA, TObjectIntMap<StatePair>> p = Automaton2Conjunction.conjunction(automatonA, automatonB, canceller);
		DFA automatonC = p.getA();
		Automaton2Minimise.minimise(automatonC, canceller);

		if (canceller.isCancelled() || p == null) {
			return Double.MIN_VALUE;
		}

		TIntObjectMap<TIntSet> automaton1State2conjunctionStates = MapConjunction2
				.reverseMapping(MapConjunction2.mapConjunction(automatonC, automatonA, canceller), canceller);

		if (canceller.isCancelled() || automaton1State2conjunctionStates == null) {
			return Double.MIN_VALUE;
		}

		int sumTransitions = 0;
		int sumTakenTransitions = 0;
		EdgeIterableOutgoing itA = automatonA.getOutgoingEdgesIterator(automatonA.getInitialState());
		EdgeIterableOutgoing itC = automatonC.getOutgoingEdgesIterator(automatonC.getInitialState());

		//walk through the states of automatonA
		BitSet statesA = DFAOperations.getReachableStates(automatonA);
		for (int stateA = statesA.nextSetBit(0); stateA >= 0; stateA = statesA.nextSetBit(stateA + 1)) {

			//walk through the mapped conjunctive states
			if (automaton1State2conjunctionStates.containsKey(stateA)) {
				for (TIntIterator it = automaton1State2conjunctionStates.get(stateA).iterator(); it.hasNext();) {
					int stateC = it.next();
					//count original
					itA.reset(stateA);
					for (itA.reset(stateA); itA.hasNext();) {
						sumTransitions++;
						itA.nextEdge();
					}

					if (automatonA.isAccepting(stateA)) {
						sumTransitions++;
					}

					//count intersect state
					itC.reset(stateC);
					for (itC.reset(stateC); itC.hasNext();) {
						sumTakenTransitions++;
						itC.nextEdge();
					}

					if (automatonC.isAccepting(stateC)) {
						sumTakenTransitions++;
					}

					if (canceller.isCancelled()) {
						return Double.MIN_VALUE;
					}
				}
			} else {
				itA.reset(stateA);
				for (itA.reset(stateA); itA.hasNext();) {
					sumTransitions++;
					itA.nextEdge();
				}

				if (automatonA.isAccepting(stateA)) {
					sumTransitions++;
				}
			}

			if (canceller.isCancelled()) {
				return Double.MIN_VALUE;
			}
		}

		if (sumTransitions != 0) {
			return sumTakenTransitions / (sumTransitions * 1.0);
		} else {
			return 1;
		}
	}
}
