package org.processmining.projectedrecallandprecision.helperclasses;

import java.util.Iterator;

/**
 * Packs several markings (byte[]) into a single byte[], while keeping them
 * sorted for easy comparison. Idea: the marking set first contains the number
 * of bytes of the first marking (in 4 bytes), then the first marking, then the
 * number of bytes of the second marking (in 4 bytes), the second marking, etc.
 * 
 * @author sander
 *
 */
public class AcceptingPetriNetMarkingSet {

	private static final int headerLength = 4;

	public static byte[] create(byte[] marking) {
		byte[] result = new byte[marking.length + headerLength];
		//header
		setHeader(result, 0, marking.length);
		//marking
		System.arraycopy(marking, 0, result, headerLength, marking.length);
		return result;
	}

	/**
	 * 
	 * @param markingSet
	 * @param marking
	 * @return the updated markingSet or the original one if the to-be-added
	 *         marking was already present.
	 */
	public static byte[] add(byte[] markingSet, byte[] marking) {
		byte[] result = new byte[markingSet.length + marking.length + headerLength];

		int currentIndex = 0;
		while (currentIndex < markingSet.length) {
			//read the header
			int currentMarkingLength = getMarkingLength(markingSet, currentIndex);
			int compare = compare(markingSet, marking, currentIndex);
			if (compare < 0) {
				//the new marking should not come before this marking; continue
				currentIndex += headerLength + currentMarkingLength;
			} else if (compare == 0) {
				//the marking is equal and was already present; do nothing
				return markingSet;
			} else {
				//the current marking should come after the to-be inserted  
				break;
			}
		}

		//the new marking should come before this marking; insert
		System.arraycopy(markingSet, 0, result, 0, currentIndex);
		//header
		setHeader(result, currentIndex, marking.length);
		System.arraycopy(marking, 0, result, currentIndex + headerLength, marking.length);
		//existing markings afterwards
		System.arraycopy(markingSet, currentIndex, result, currentIndex + headerLength + marking.length,
				markingSet.length - currentIndex);

		return result;
	}

	public static byte[] addAll(byte[] markingSetA, byte[] markingSetB) {
		byte[] result = markingSetA;
		for (Iterator<byte[]> it = getMarkings(markingSetB); it.hasNext();) {
			byte[] marking = it.next();
			result = add(result, marking);
		}
		return result;
	}

	public static Iterator<byte[]> getMarkings(final byte[] markingSet) {
		return new Iterator<byte[]>() {
			int now = -1;
			int next = 0;

			public boolean hasNext() {
				return next < markingSet.length;
			}

			public byte[] next() {
				now = next;
				int markingLength = getMarkingLength(markingSet, now);
				next += headerLength + markingLength;
				byte[] result = new byte[markingLength];
				System.arraycopy(markingSet, now + headerLength, result, 0, markingLength);
				return result;
			}
			
			public void remove() {
				//not implemented
			}
		};
	}

	public static boolean contains(byte[] markingSet, byte[] marking) {
		int currentIndex = 0;
		while (currentIndex < markingSet.length) {
			//read the header
			int currentMarkingLength = getMarkingLength(markingSet, currentIndex);
			int compare = compare(markingSet, marking, currentIndex);
			if (compare < 0) {
				//the new marking should not come before this marking; continue
				currentIndex += headerLength + currentMarkingLength;
			} else if (compare == 0) {
				//the marking is equal and was already present
				return true;
			} else {
				//we are past the point where this marking should occur
				return false;
			}
		}
		return false;
	}

	private static int compare(byte[] markingSet, byte[] markingA, int headerIndexB) {
		int lengthB = getMarkingLength(markingSet, headerIndexB);
		if (markingA.length < lengthB) {
			return 1;
		} else if (markingA.length > lengthB) {
			return -1;
		}

		int indexA = 0;
		for (int indexB = headerIndexB + headerLength; indexB < headerIndexB + headerLength + lengthB; indexB++) {
			if (markingA[indexA] < markingSet[indexB]) {
				return 1;
			} else if (markingA[indexA] > markingSet[indexB]) {
				return -1;
			}
			indexA++;
		}

		return 0;
	}

	private static int getMarkingLength(byte[] markingSet, int headerIndex) {
		return markingSet[headerIndex] << 24 | (markingSet[headerIndex + 1] & 0xFF) << 16
				| (markingSet[headerIndex + 2] & 0xFF) << 8 | (markingSet[headerIndex + 3] & 0xFF);
	}

	private static void setHeader(byte[] markingSet, int headerIndex, int length) {
		markingSet[headerIndex] = (byte) (length >> 24);
		markingSet[headerIndex + 1] = (byte) (length >> 16);
		markingSet[headerIndex + 2] = (byte) (length >> 8);
		markingSet[headerIndex + 3] = (byte) length;
	}

	/**
	 * 
	 * @param markingSet
	 * @param finalMarkings
	 * @return whether at least one of the markings in the markingSet is a final
	 *         marking.
	 */
	public static boolean containsFinalMarking(byte[] markingSet, Iterable<byte[]> finalMarkings) {
		for (byte[] finalMarking : finalMarkings) {
			int currentIndex = 0;
			while (currentIndex < markingSet.length) {
				//read the header
				int currentMarkingLength = getMarkingLength(markingSet, currentIndex);
				int compare = compare(markingSet, finalMarking, currentIndex);
				if (compare < 0) {
					//the new marking should not come before this marking; continue
					currentIndex += headerLength + currentMarkingLength;
				} else if (compare == 0) {
					//the marking is equal and was already present
					return true;
				} else {
					break;
				}
			}
		}
		return false;
	}
}
