package org.processmining.projectedrecallandprecision.helperclasses;

import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTree;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTreeImpl;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTreeReduce;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTreeReduce.ReductionFailedException;
import org.processmining.plugins.InductiveMiner.efficienttree.UnknownTreeNodeException;
import org.processmining.projectedrecallandprecision.framework.CompareParameters;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;

import gnu.trove.map.TObjectIntMap;

public class EfficientTreeProject {

	/**
	 * Projects a tree and transforms it into an automaton
	 * 
	 * @param system
	 * @param canceller
	 * @param names
	 * @return
	 * @throws ReductionFailedException
	 * @throws UnknownTreeNodeException
	 * @throws CloneNotSupportedException
	 */
	public static DFA projectAndAutomatiseTree(EfficientTree system, CompareParameters parameters,
			ProMCanceller canceller, String... names)
			throws ReductionFailedException, UnknownTreeNodeException, CloneNotSupportedException {
		//project and reduce
		EfficientTree systemProjected = EfficientTreeProject.project(system, names);
		EfficientTreeReduce.reduce(systemProjected, parameters.getTreeReduceParameters());

		//transform to automaton
		DFA automaton = Tree2Automaton.convert(systemProjected, canceller);

		if (canceller.isCancelled() || automaton == null) {
			return null;
		}

		return automaton;
	}

	/**
	 * Project a tree onto activities.
	 * 
	 * @param efficientTree
	 * @param activity2short
	 * @param keepActivities
	 * @return
	 * @throws CloneNotSupportedException
	 */
	public static EfficientTree project(EfficientTree efficientTree, String... keepActivities)
			throws CloneNotSupportedException {
		//map: [newIndex] = oldindex
		short[] keepActivitiesShort = new short[keepActivities.length];

		TObjectIntMap<String> newActivity2short = EfficientTreeImpl.getEmptyActivity2int();
		String[] newShort2activity = new String[keepActivities.length];

		for (short newIndex = 0; newIndex < keepActivities.length; newIndex++) {
			String activity = keepActivities[newIndex];
			int oldIndex = efficientTree.getActivity2int().get(activity);

			//if the activity does not appear in the tree, oldIndex will be < 0.
			assert (oldIndex < Short.MAX_VALUE);
			keepActivitiesShort[newIndex] = (short) oldIndex;

			newActivity2short.put(activity, newIndex);
			newShort2activity[newIndex] = activity;
		}

		return project(efficientTree, keepActivitiesShort);
	}

	/**
	 * 
	 * @param efficientTree
	 * @param keepActivities
	 * @return a copy of the tree in which every activity not
	 *         from @keepActivities is replaced with tau, and every occurrence
	 *         is replaced by the number of the activity as given in
	 *         keepActivities.
	 * @throws CloneNotSupportedException
	 */
	public static EfficientTree project(EfficientTree efficientTree, short... keepActivities)
			throws CloneNotSupportedException {
		EfficientTree result = efficientTree.clone();

		//walk through the tree and replace things
		for (int i = 0; i < result.getMaxNumberOfNodes(); i++) {
			if (result.getActivity(i) >= 0) {
				boolean found = false;
				for (short j = 0; j < keepActivities.length; j++) {
					if (result.getActivity(i) == keepActivities[j]) {
						result.setNodeActivity(i, j);
						found = true;
						break;
					}
				}
				if (!found) {
					result.setNodeType(i, EfficientTree.NodeType.tau);
				}
			}
		}

		return result;
	}

}
