package org.processmining.projectedrecallandprecision.helperclasses;

import java.lang.reflect.Array;
import java.util.List;
import java.util.concurrent.atomic.AtomicLong;

import org.processmining.plugins.InductiveMiner.Pair;

/**
 * Iterate over all cartesian product combinations in a thread-safe way
 * 
 * @author sleemans
 *
 */
public class ExponentialIteratorThreadSafe<X> {

	//input list
	final List<X> list;
	final Class<X> clazz;

	//number of times cartesian product
	final int cardinality;

	//buckets
	final AtomicLong[] bucketIsAt;
	final long[] bucketEndsAt;

	//progress keeping
	final AtomicLong progress;

	public ExponentialIteratorThreadSafe(Class<X> clazz, List<X> list, int cardinality, int numberOfThreads) {

		this.list = list;
		this.clazz = clazz;
		this.cardinality = cardinality;
		progress = new AtomicLong(0);

		long totalNumberOfPairs = (long) Math.pow(list.size(), cardinality);

		//initialise buckets
		bucketIsAt = new AtomicLong[numberOfThreads];
		bucketEndsAt = new long[numberOfThreads];
		for (int bucket = 0; bucket < numberOfThreads; bucket++) {
			bucketIsAt[bucket] = new AtomicLong(bucket * (totalNumberOfPairs / numberOfThreads));
			bucketEndsAt[bucket] = (bucket + 1) * (totalNumberOfPairs / numberOfThreads) - 1;
		}
		bucketEndsAt[numberOfThreads - 1] = totalNumberOfPairs - 1;
	}

	/**
	 * 
	 * @param threadNr
	 * @return the number of the combination that is to be processed next.
	 *         Guaranteed to be thread safe. Returns null if all combinations
	 *         have been processed.
	 */
	public X[] next(int threadNr) {

		long p = getNextPair(threadNr);
		if (p == -1) {
			return null;
		}

		@SuppressWarnings("unchecked")
		X[] result = (X[]) Array.newInstance(clazz, cardinality);

		for (int i = 0; i < cardinality - 1; i++) {
			result[i] = list.get((int) (p % list.size()));
			p = p / list.size();
		}
		result[cardinality - 1] = list.get((int) p);

		//keep track of progress
		{
			long x = progress.incrementAndGet();
			long max = bucketEndsAt[bucketEndsAt.length - 1] + 1;
			if (x % 10000000 == 0) {
				System.out.println("start combination " + x + " of " + max);
			}
		}

		return result;
	}

	private long getNextPair(int threadNr) {
		int threads = bucketIsAt.length;
		for (int t = threadNr; t != ((threadNr - 1) + threads) % threads; t = (t + 1) % threads) {
			long pair = bucketIsAt[t].getAndIncrement();
			if (pair <= bucketEndsAt[t]) {
				return pair;
			}
		}
		return -1;
	}

	/**
	 * Put the combinations in buckets to spread computation.
	 * 
	 * @param numberOfThreads
	 * @param numberOfPairs
	 * @return
	 */
	public static Pair<AtomicLong[], long[]> initNextPair(int numberOfThreads, long numberOfPairs) {
		final AtomicLong[] atPairs = new AtomicLong[numberOfThreads];
		final long[] atPairMax = new long[numberOfThreads];

		for (int bucket = 0; bucket < numberOfThreads; bucket++) {
			atPairs[bucket] = new AtomicLong(bucket * (numberOfPairs / numberOfThreads));
			atPairMax[bucket] = (bucket + 1) * (numberOfPairs / numberOfThreads) - 1;
		}
		atPairMax[numberOfThreads - 1] = numberOfPairs - 1;

		return Pair.of(atPairs, atPairMax);
	}

}
