package org.processmining.projectedrecallandprecision.helperclasses;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import org.apache.commons.collections4.iterators.PermutationIterator;
import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.plugins.InductiveMiner.efficienttree.EfficientTree;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAFactory;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Concatenate;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Minimise;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Repeat;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Shuffle;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.Automaton2Union;

public class Tree2Automaton {

	public static DFA convert(EfficientTree tree, ProMCanceller canceller) {
		DFA automaton = convert(tree, 0, canceller);

		if (canceller.isCancelled() || automaton == null) {
			return null;
		}

		Automaton2Minimise.minimise(automaton, canceller);
		if (canceller.isCancelled()) {
			return null;
		}

		return automaton;
	}

	private static DFA convert(EfficientTree tree, int node, ProMCanceller canceller) {

		if (canceller.isCancelled()) {
			return null;
		}

		if (tree.isTau(node)) {
			DFA result = DFAFactory.createAutomaton();
			result.setAccepting(result.getInitialState(), true);
			return result;
		} else if (tree.isActivity(node)) {
			DFA result = DFAFactory.createAutomaton();
			int state = result.addEdge(result.getInitialState(), (short) tree.getActivity(node));
			result.setAccepting(state, true);
			return result;
		} else if (tree.isOperator(node)) {
			//make automata for children
			int numberOfChildren = tree.getNumberOfChildren(node);
			List<DFA> childrenAutomata = new ArrayList<>(numberOfChildren);
			for (int child : tree.getChildren(node)) {
				DFA childAutomaton = convert(tree, child, canceller);

				if (canceller.isCancelled() || childAutomaton == null) {
					return null;
				}

				childrenAutomata.add(childAutomaton);
			}

			if (tree.isSequence(node)) {
				return Automaton2Concatenate.concatenate(childrenAutomata, canceller);
			} else if (tree.isXor(node)) {
				return Automaton2Union.union(childrenAutomata, canceller);
			} else if (tree.isConcurrent(node)) {
				return Automaton2Shuffle.shuffle(childrenAutomata, canceller);
			} else if (tree.isLoop(node)) {
				DFA body = childrenAutomata.get(0);
				DFA redo = childrenAutomata.get(1);
				DFA exit = childrenAutomata.get(2);
				DFA repeat = Automaton2Concatenate.concatenate(redo, body, canceller);
				if (canceller.isCancelled() || repeat == null) {
					return null;
				}
				DFA star = Automaton2Repeat.repeat(repeat, canceller);
				if (canceller.isCancelled() || star == null) {
					return null;
				}
				return Automaton2Concatenate.concatenate(Arrays.asList(new DFA[] { body, star, exit }),
						canceller);
			} else if (tree.isOr(node)) {
				DFA result = getOr(childrenAutomata.get(0), childrenAutomata.get(1), canceller);
				if (canceller.isCancelled() || result == null) {
					return null;
				}
				for (int i = 2; i < numberOfChildren; i++) {
					result = getOr(result, childrenAutomata.get(i), canceller);
					if (canceller.isCancelled() || result == null) {
						return null;
					}
				}
				return result;
			} else if (tree.isInterleaved(node)) {
				PermutationIterator<DFA> it = new PermutationIterator<>(childrenAutomata);
				DFA result = Automaton2Concatenate.concatenate(it.next(), canceller);
				if (canceller.isCancelled() || result == null) {
					return null;
				}
				for (; it.hasNext();) {
					DFA result2 = Automaton2Concatenate.concatenate(it.next(), canceller);
					if (canceller.isCancelled() || result2 == null) {
						return null;
					}
					result = Automaton2Union.union(result, result2, canceller);
					if (canceller.isCancelled() || result == null) {
						return null;
					}
				}
				return result;
			}
		}
		throw new RuntimeException("not implemented");
	}

	public static DFA getOr(DFA automaton1, DFA automaton2, ProMCanceller canceller) {
		DFA a1Ua2 = Automaton2Union.union(automaton1, automaton2, canceller);
		if (canceller.isCancelled() || a1Ua2 == null) {
			return null;
		}
		DFA shu = Automaton2Shuffle.shuffle(automaton1, automaton2, canceller);
		if (canceller.isCancelled() || shu == null) {
			return null;
		}
		DFA result = Automaton2Union.union(a1Ua2, shu, canceller);

		if (canceller.isCancelled() || result == null) {
			return null;
		}

		return result;
	}
}
