package org.processmining.projectedrecallandprecision.helperclasses.automata.operations;

import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList.EdgeIterableOutgoing;
import org.processmining.projectedrecallandprecision.helperclasses.automata.operations.StatePairSet.StatePairIterator;

import gnu.trove.iterator.TShortObjectIterator;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.TShortObjectMap;
import gnu.trove.map.custom_hash.TObjectIntCustomHashMap;
import gnu.trove.map.hash.TShortObjectHashMap;
import gnu.trove.strategy.HashingStrategy;

public class Automaton2Concatenate {
	/**
	 * Returns an automaton that accepts the concatenation of the given
	 * automata. Does not modify the automata.
	 * 
	 * @param canceller
	 * @param automata
	 * @return the first automaton for convenience.
	 */
	public static DFA concatenate(List<DFA> automata, ProMCanceller canceller) {
		Iterator<DFA> it = automata.iterator();
		if (!it.hasNext()) {
			return null;
		}

		DFA result = it.next();

		while (it.hasNext()) {
			result = concatenate(result, it.next(), canceller);

			if (result == null || canceller.isCancelled()) {
				return null;
			}
		}

		return result;
	}

	public static DFA concatenate(DFA automatonA, DFA automatonB, ProMCanceller canceller) {
		DFA result = new DFAImplEdgeList();

		ArrayDeque<int[]> worklist = new ArrayDeque<>();
		final TObjectIntMap<int[]> statePairSet2resultState = new TObjectIntCustomHashMap<>(
				new HashingStrategy<int[]>() {
					private static final long serialVersionUID = -4818184406522740901L;

					public int computeHashCode(int[] object) {
						return Arrays.hashCode(object);
					}

					public boolean equals(int[] o1, int[] o2) {
						return Arrays.equals(o1, o2);
					}
				}, 10, 0.5f, -1);
		{
			int[] statePair = StatePairSet.create(automatonA.getInitialState(),
					automatonA.isAccepting(automatonA.getInitialState()) ? automatonB.getInitialState() : -1);
			worklist.add(statePair);
			statePairSet2resultState.put(statePair, result.getInitialState());
		}

		EdgeIterableOutgoing itA = automatonA.getOutgoingEdgesIterator(automatonA.getInitialState());
		EdgeIterableOutgoing itB = automatonB.getOutgoingEdgesIterator(automatonB.getInitialState());
		StatePairIterator itStates = new StatePairSet.StatePairIterator();

		while (!worklist.isEmpty()) {
			int[] sourcePairSet = worklist.removeFirst();
			int resultSourceStateSet = statePairSet2resultState.get(sourcePairSet);

			//set the state accepting if necessary
			result.setAccepting(resultSourceStateSet, StatePairSet.oneStateBIsAccepting(automatonB, sourcePairSet));

			//gather the next states
			TShortObjectMap<int[]> nextStates = new TShortObjectHashMap<>(10, 0.5f, (short) -1);

			for (itStates.reset(sourcePairSet); itStates.hasNext();) {
				itStates.next();

				//walk through the edges of nodeA; add a possibility for all outgoing edges.
				int sourceStateA = itStates.getNodeA();
				int sourceStateB = itStates.getNodeB();

				//add the possible states going from this stateA in this statePair
				if (sourceStateA != -1) {
					for (itA.reset(sourceStateA); itA.hasNext();) {
						itA.nextEdge();

						int targetStateA = itA.getTarget();
						int targetStateB = sourceStateB;
						//perform the magic: if A reaches an end state, that's equivalent to the initial state of B.
						if (targetStateB == -1 && automatonA.isAccepting(targetStateA)) {
							targetStateB = automatonB.getInitialState();
						}

						short activity = itA.getActivity();

						if (nextStates.containsKey(activity)) {
							nextStates.put(activity,
									StatePairSet.add(nextStates.get(activity), targetStateA, targetStateB));
						} else {
							nextStates.put(activity, StatePairSet.create(targetStateA, targetStateB));
						}
					}
				}

				//add the possible states going from this stateB in this statePair
				if (sourceStateB != -1) {
					for (itB.reset(sourceStateB); itB.hasNext();) {
						itB.nextEdge();

						/*
						 * Perform the magic: once a step in B is made, we are
						 * not in A anymore.
						 */
						int targetStateA = -1;
						int targetStateB = itB.getTarget();
						short activity = itB.getActivity();

						if (nextStates.containsKey(activity)) {
							nextStates.put(activity,
									StatePairSet.add(nextStates.get(activity), targetStateA, targetStateB));
						} else {
							nextStates.put(activity, StatePairSet.create(targetStateA, targetStateB));
						}
					}
				}
			}

			for (TShortObjectIterator<int[]> it = nextStates.iterator(); it.hasNext();) {
				it.advance();
				int[] statePairSetTarget = it.value();
				short activity = it.key();
				int resultTargetState = statePairSet2resultState.get(statePairSetTarget);
				if (resultTargetState == statePairSet2resultState.getNoEntryValue()) {
					//this state pair did not exist yet
					resultTargetState = result.addEdge(resultSourceStateSet, activity);
					statePairSet2resultState.put(statePairSetTarget, resultTargetState);

					worklist.add(statePairSetTarget);
				} else {
					//this state pair did not exist yet
					result.addEdge(resultSourceStateSet, activity, resultTargetState);
				}
			}
		}

		return result;
	}
}
