package org.processmining.projectedrecallandprecision.helperclasses.automata.operations;

import java.util.ArrayDeque;

import org.processmining.framework.plugin.ProMCanceller;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFA;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAFactory;
import org.processmining.projectedrecallandprecision.helperclasses.automata.DFAImplEdgeList.EdgeIterableOutgoing;

import gnu.trove.iterator.TIntIterator;
import gnu.trove.iterator.TShortObjectIterator;
import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.TShortObjectMap;
import gnu.trove.map.hash.TObjectIntHashMap;
import gnu.trove.map.hash.TShortObjectHashMap;
import gnu.trove.set.TIntSet;
import gnu.trove.set.hash.TIntHashSet;

public class Automaton2Repeat {
	/**
	 * Constructs a new automaton that accepts any repetition (Kleene star) of
	 * the original automaton.
	 * 
	 * @param canceller
	 * @param automatonA
	 * @return the automaton for convenience.
	 */
	public static DFA repeat(DFA automatonA, ProMCanceller canceller) {
		DFA result = DFAFactory.createAutomaton();

		ArrayDeque<TIntSet> worklist = new ArrayDeque<>();
		TObjectIntMap<TIntSet> stateSet2resultState = new TObjectIntHashMap<>(10, 0.5f, -1);
		{
			TIntSet initialStateSet = new TIntHashSet(1, 1, -1);
			initialStateSet.add(result.getInitialState());
			worklist.add(initialStateSet);
			stateSet2resultState.put(initialStateSet, result.getInitialState());
		}

		EdgeIterableOutgoing itA = automatonA.getOutgoingEdgesIterator(automatonA.getInitialState());

		while (!worklist.isEmpty()) {

			TIntSet sourceStateSet = worklist.removeFirst();
			int resultSourceState = stateSet2resultState.get(sourceStateSet);

			//set the state accepting if necessary
			result.setAccepting(resultSourceState, isAnyAccepting(automatonA, sourceStateSet));

			//gather outgoing edges and next states
			TShortObjectMap<TIntSet> nextStates = new TShortObjectHashMap<>(10, 0.5f, (short) -1);
			{
				for (TIntIterator it = sourceStateSet.iterator(); it.hasNext();) {
					int stateA = it.next();

					//walk through all outgoing edges of stateA
					for (itA.reset(stateA); itA.hasNext();) {
						itA.nextEdge();

						int targetStateA = itA.getTarget();
						short activity = itA.getActivity();

						TIntSet old;
						if (nextStates.containsKey(activity)) {
							old = nextStates.get(activity);
						} else {
							old = new TIntHashSet(10, 0.5f, -1);
						}
						old.add(targetStateA);

						/*
						 * Perform the magic: if the targetStateA is accepting,
						 * then we can make an "epsilon" move to the initial
						 * state.
						 */
						if (automatonA.isAccepting(targetStateA)) {
							old.add(result.getInitialState());
						}

						nextStates.put(activity, old);
					}
				}
			}

			//add states to result automaton
			for (TShortObjectIterator<TIntSet> it = nextStates.iterator(); it.hasNext();) {
				it.advance();
				TIntSet stateSetTarget = it.value();
				short activity = it.key();
				int resultTargetState = stateSet2resultState.get(stateSetTarget);
				if (resultTargetState == stateSet2resultState.getNoEntryValue()) {
					//this state pair did not exist yet
					resultTargetState = result.addEdge(resultSourceState, activity);
					stateSet2resultState.put(stateSetTarget, resultTargetState);

					worklist.add(stateSetTarget);
				} else {
					//this state pair did not exist yet
					result.addEdge(resultSourceState, activity, resultTargetState);
				}
			}

			if (canceller.isCancelled()) {
				return null;
			}
		}

		return result;
	}

	public static boolean isAnyAccepting(DFA automaton, TIntSet stateSet) {
		for (TIntIterator it = stateSet.iterator(); it.hasNext();) {
			int state = it.next();
			if (automaton.isAccepting(state) || state == automaton.getInitialState()) {
				return true;
			}
		}
		return false;
	}
}
