package org.processmining.projectedrecallandprecision.result;

import java.util.Arrays;
import java.util.Collection;

import org.apache.commons.lang3.ArrayUtils;
import org.processmining.projectedrecallandprecision.framework.CompareParameters.RecallName;
import org.processmining.projectedrecallandprecision.helperclasses.AutomatonFailedException;

import gnu.trove.map.TObjectIntMap;
import gnu.trove.map.custom_hash.TObjectIntCustomHashMap;
import gnu.trove.strategy.HashingStrategy;

public class ProjectedRecallPrecisionResultImplBasic implements ProjectedRecallPrecisionResult {
	private final int k;
	private final String[] int2activity;
	private final TObjectIntMap<String> activity2int;
	private final double[] recall;
	private final double[] precision;

	private final boolean recallComputed;
	private final boolean precisionComputed;

	private final Object system;
	private final Object model;

	private final RecallName recallName;

	public ProjectedRecallPrecisionResultImplBasic(int k, Collection<String> activities, RecallName recallName,
			Object system, Object model, boolean recallComputed, boolean precisionComputed)
			throws AutomatonFailedException {
		this.k = k;
		this.recallName = recallName;
		this.system = system;
		this.model = model;
		this.recallComputed = recallComputed;
		this.precisionComputed = precisionComputed;

		//copy and sort the activities
		{
			int2activity = new String[activities.size()];
			int i = 0;
			for (String activity : activities) {
				int2activity[i] = activity;
				i++;
			}
			Arrays.sort(int2activity, comparator);
			activities = null;
		}

		//initialise the hashmap. Note the no-entry-value being negative to avoid collisions'
		{
			activity2int = new TObjectIntCustomHashMap<String>(new HashingStrategy<String>() {
				private static final long serialVersionUID = 8085536417507375890L;

				public int computeHashCode(String object) {
					return object.hashCode();
				}

				public boolean equals(String o1, String o2) {
					return o1.equals(o2);
				}
			}, 10, 0.5f, -1);
			for (int i = 0; i < int2activity.length; i++) {
				activity2int.put(int2activity[i], i);
			}
		}

		//initialise recall and precision to -1
		long numberOfSubsets = binom(int2activity.length, k);
		if (numberOfSubsets > Integer.MAX_VALUE) {
			throw new AutomatonFailedException("too many projected subsets to hold in an array");
		}
		recall = new double[(int) numberOfSubsets];
		precision = new double[(int) numberOfSubsets];

		for (int i = 0; i < numberOfSubsets; i++) {
			recall[i] = -1;
			precision[i] = -1;
		}
	}

	/**
	 * 
	 * @param recall
	 * @param precision
	 * @param activities
	 *            Thread safe on unique activities
	 */
	public void set(double recall, double precision, String... activities) {
		assert (activities.length == k);
		int index = getIndex(activities);
		this.recall[index] = recall;
		this.precision[index] = precision;

		//System.out.println("$\\{" + String.join(", ", copyAndSort(activities)) + "\\}$&" + new DecimalFormat("0.000").format(recall)
		//		+ "&" + new DecimalFormat("0.000").format(precision) + "\\\\");
	}

	private int getIndex(String... activities) {
		//copy and sort the activities
		String[] activities2 = copyAndSort(activities);

		//get start of the row
		int index = 0;
		for (int i = activities2.length; i > 0; i--) {
			index += (int) binom(activity2int.get(activities2[i - 1]), i);
		}

		return index;
	}

	/**
	 * 
	 * @param activities
	 *            Can be of any size <= k.
	 * @return the (average) recall over all measurements performed for the
	 *         given number of activities.
	 * @throws ProjectedMeasuresFailedException
	 */
	public double getRecall(String... activities) throws ProjectedMeasuresFailedException {
		if (getNumberOfFailedMeasures() > 0) {
			throw new ProjectedMeasuresFailedException();
		}
		return getMeasure(recall, activities);
	}

	/**
	 * 
	 * @param activities
	 *            Can be of any size <= k.
	 * @return the (average) precision over all measurements performed for the
	 *         given number of activities.
	 * @throws ProjectedMeasuresFailedException
	 */
	public double getPrecision(String... activities) throws ProjectedMeasuresFailedException {
		if (getNumberOfFailedMeasures() > 0) {
			throw new ProjectedMeasuresFailedException();
		}
		return getMeasure(precision, activities);
	}

	private double getMeasure(double[] measures, String... activities) throws ProjectedMeasuresFailedException {
		if (getNumberOfFailedMeasures() > 0) {
			throw new ProjectedMeasuresFailedException();
		}
		assert activities.length <= k;
		if (activities.length == 0 || activities.length == k) {
			//compute average over all measurements
			double sum = 0;
			int count = 0;
			for (int i = 0; i < measures.length; i++) {
				if (measures[i] >= -0.01) {
					count++;
					sum += measures[i];
				}
			}
			if (count == 0) {
				//if nothing was measured, the measure is perfect
				return 1;
			}
			return sum / (count * 1.0);
		} else if (activities.length == k) {
			//get a single subset
			return measures[getIndex(activities)];
		} else {
			//compute average over one dimension lower
			String[] activities2 = new String[activities.length + 1];
			System.arraycopy(activities, 0, activities2, 1, activities.length);

			double sum = 0;
			int count = 0;
			for (int i = 0; i < int2activity.length; i++) {
				activities2[0] = int2activity[i];
				if (!ArrayUtils.contains(activities, int2activity[i])) {
					double m = getMeasure(measures, activities2);
					if (m >= -0.01) {
						sum += m;
						count++;
					}
				}
			}
			if (count == 0) {
				//if nothing was measured, the measure was perfect
				return 1;
			}
			return sum / (count * 1.0);
		}
	}

	/**
	 * 
	 * @return the number of measures that failed
	 */
	public int getNumberOfFailedMeasures() {
		int count = 0;
		for (int i = 0; i < recall.length; i++) {
			if (recall[i] < -1.5) {
				count++;
			}
		}
		return count;
	}

	private String[] copyAndSort(String[] array) {
		String[] result = new String[array.length];
		System.arraycopy(array, 0, result, 0, array.length);
		Arrays.sort(result, comparator);
		return result;
	}

	public int getK() {
		return k;
	}

	public RecallName getRecallName() {
		return recallName;
	}

	public String[] getActivities() {
		return int2activity;
	}

	public long binom(int n, int k) {
		if (n < k) {
			return 0;
		}
		long[] b = new long[n + 1];
		b[0] = 1;
		for (int i = 1; i <= n; i++) {
			b[i] = 1;
			for (int j = i - 1; j > 0; --j) {
				b[j] += b[j - 1];
			}
		}
		return b[k];
	}

	public Object getSystem() {
		return system;
	}

	public Object getModel() {
		return model;
	}

	public boolean isRecallComputed() {
		return recallComputed;
	}

	public boolean isPrecisionComputed() {
		return precisionComputed;
	}

}